import errno
import os
import logging

from contextlib import ContextDecorator, contextmanager
from spacemake.errors import SpacemakeError, FileWrongExtensionError

LINE_SEPARATOR = "-" * 50 + "\n"

bool_in_str = ["True", "true", "False", "false"]

def assert_file(file_path, default_value=None, extension=["all"]):
    if file_path == default_value:
        # file doesn't exist but has the default value,
        # so we do not need to assert anything
        return False

    if isinstance(extension, str):
        extension = [extension]

    if not isinstance(file_path, list):
        file_path = [file_path]

    for fp in file_path:
        # check if file exists, raise error if not
        if not os.path.isfile(fp):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fp)

        # check for all extensions
        if extension != ["all"]:
            extension_check = [fp.endswith(ex) for ex in extension]
            if not any(extension_check):
                raise FileWrongExtensionError(fp, extension)

    # return true if file exists and every test was good
    return True


def str2bool(var):
    if isinstance(var, bool):
        return var

    if var in ["True", "true"]:
        return True
    elif var in ["False", "false"]:
        return False
    else:
        raise ValueError(f"variable should be boolean, or one of: {bool_in_str}")


def ensure_path(path):
    import os


def ensure_path(path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return path


def FASTQ_src(src):
    from more_itertools import grouper

    for name, seq, _, qual in grouper(src, 4):
        yield name.rstrip()[1:], seq.rstrip(), qual.rstrip()


def BAM_src(src):
    import pysam

    bam = pysam.AlignmentFile(src, "rb", check_sq=False)
    for read in bam.fetch(until_eof=True):
        yield read.query_name, read.query_sequence, read.query_qualities


def read_fq(fname):
    import gzip

    if fname.endswith(".gz"):
        src = FASTQ_src(gzip.open(fname, mode="rt"))
    elif fname.endswith(".bam"):
        src = BAM_src(fname)
    elif type(fname) is str:
        src = FASTQ_src(open(fname))
    else:
        src = FASTQ_src(fname)  # assume its a stream or file-like object already

    for name, seq, qual in src:
        yield name, seq, qual


def dge_to_sparse(dge_path):
    import anndata
    import numpy as np
    import gzip
    import pandas as pd
    from scipy.sparse import csr_matrix, vstack

    ix = 0
    mix = 0
    matrices = []
    gene_names = []

    with gzip.open(dge_path, "rt") as dge:
        first_line = dge.readline().strip().split("\t")

        barcodes = first_line[1:]
        ncol = len(barcodes)
        M = np.zeros((1000, ncol))

        for line in dge:
            vals = line.strip().split("\t")
            # first element contains the gene name
            gene_names.append(vals[0])
            vals = vals[1:]
            vals = np.array(vals, dtype=np.int64)

            M[ix] = vals

            if ix % 1000 == 999:
                print(mix)
                mix = mix + 1
                matrices.append(csr_matrix(M))
                ix = 0
                M = np.zeros((1000, ncol))
            else:
                ix = ix + 1

        # get the leftovers
        M = M[:ix]
        matrices.append(csr_matrix(M))

        # sparse expression matrix
        X = vstack(matrices, format="csr")
        print(len(gene_names))

        adata = anndata.AnnData(
            X.T, obs=pd.DataFrame(index=barcodes), var=pd.DataFrame(index=gene_names)
        )

        return adata


def compute_neighbors(adata, min_dist=None, max_dist=None):
    """Compute all direct neighbors of all spots in the adata. Currently
        tailored for 10X Visium.
    Args:
        adata: an AnnData object
        min_dist: int, minimum distance to consider neighbors
        max_dist: int, maximum distance to consider neighbors
    Returns:
        neighbors: a dictionary holding the spot IDs for every spot
    """
    import numpy as np
    from scipy.spatial.distance import cdist

    # Calculate all Euclidean distances on the adata
    dist_mtrx = cdist(adata.obsm["spatial"], adata.obsm["spatial"])
    neighbors = dict()

    for i in range(adata.obs.shape[0]):
        neighbors[i] = np.where(
            (dist_mtrx[i, :] > min_dist) & (dist_mtrx[i, :] < max_dist)
        )[0]

    return neighbors


def compute_islands(adata, min_umi):
    """Find contiguity islands

    Args:
        adata: an AnnData object
        cluster: a cell type label to compute the islands for
    Returns:
        islands: A list of lists of spots forming contiguity islands
    """
    import numpy as np
    import itertools

    # this is hard coded for now for visium, to have 6 neighbors per spot
    # TODO: define an iterative approach where the key is to have around 6
    # neighbors per spot on average
    neighbors = compute_neighbors(adata, min_dist=0, max_dist=3)
    spots_cluster = np.where(np.array(adata.obs["total_counts"]) < min_umi)[0]

    # create a new dict holding only neighbors of the cluster
    islands = []

    for spot in neighbors:
        if spot not in spots_cluster:
            continue
        islands.append({spot}.union({x for x in neighbors[spot] if x in spots_cluster}))

    # merge islands with common spots
    island_spots = set(itertools.chain.from_iterable(islands))

    for each in island_spots:
        components = [x for x in islands if each in x]
        for i in components:
            islands.remove(i)
        islands += [list(set(itertools.chain.from_iterable(components)))]

    return islands


def detect_tissue(adata, min_umi):
    """Detect tissue: first find beads with at least min_umi UMIs, then detect island in the rest

    Args
        adata: an AnnData object, with spatial coordinates
        min_umi: integer, the min umi to be assigned as tissue bead by default
    Returns:
        tissue_indices: a list of indices which should be kept for this AnnData object
    """
    import numpy as np

    islands = compute_islands(adata, min_umi)

    # find the sizes of the islands. remove the biggest, as either the tissue has a big hole in it
    # or there are not so many big islands in which case removal is OK.
    # to be evaluated later...
    island_sizes = [len(island) for island in islands]

    tissue_islands = np.delete(islands, np.argmax(island_sizes))

    # get the indices of the islands
    tissue_indices = np.where(np.array(adata.obs["total_counts"]) >= min_umi)[0]

    tissue_indices = np.append(tissue_indices, np.hstack(tissue_islands))

    return tissue_indices


COMPLEMENT = {
    "a": "t",
    "t": "a",
    "c": "g",
    "g": "c",
    "k": "m",
    "m": "k",
    "r": "y",
    "y": "r",
    "s": "s",
    "w": "w",
    "b": "v",
    "v": "b",
    "h": "d",
    "d": "h",
    "n": "n",
    "A": "T",
    "T": "A",
    "C": "G",
    "G": "C",
    "K": "M",
    "M": "K",
    "R": "Y",
    "Y": "R",
    "S": "S",
    "W": "W",
    "B": "V",
    "V": "B",
    "H": "D",
    "D": "H",
    "N": "N",
    "-": "-",
    "=": "=",
    "+": "+",
}


def complement(s):
    return "".join([COMPLEMENT[x] for x in s])


def rev_comp(seq):
    return complement(seq)[::-1]


def fasta_chunks(lines, strip=True, fuse=True):
    chunk = ""
    data = []

    for l in lines:
        if l.startswith("#"):
            continue
        if l.startswith(">"):
            if data and chunk:
                # print chunk
                yield chunk, "".join(data)

                if strip:
                    data = []
                else:
                    data = [l]

            chunk = l[1:].strip()
        else:
            if fuse:
                data.append(l.rstrip())
            else:
                data.append(l)

    if data and chunk:
        yield chunk, "".join(data)


@contextmanager
def message_aggregation(
    log_listen="spacemake",
    print_logger=False,
    print_success=True
):
    message_buffer = []

    log = logging.getLogger(log_listen)
    log.setLevel(logging.INFO)

    class MessageHandler(logging.NullHandler):
        def handle(this, record):
            if record.name == log_listen:
                if print_logger:
                    print(f"{log_listen}: {record.msg}")
                else:
                    print(record.msg)

    log.addHandler(MessageHandler())

    try:
        yield True

        if print_success:
            print(f"{LINE_SEPARATOR}SUCCESS!")

    except SpacemakeError as e:
        print(e)

def str_to_list(value):
    # if list in string representation, return the list
    if value is None:
        return None

    if type(value) is str and value.startswith("["):
        if value == "[nan]":
            return []
        else:
            return eval(value)
    # else create a list
    else:
        return [value]

def check_star_index_compatibility(star_index_dir):
    import os
    
    star_version = os.popen('STAR --version').read().strip()

    with open(os.path.join(star_index_dir, 'Log.out'), 'r') as f:
        first_line = f.readline().strip()
        index_version = first_line.split('=')[-1].split('_')[-1]

        if index_version != star_version:
            raise SpacemakeError(f'STAR index version ({index_version}) is' +
                f' incompatible with your STAR version ({star_version})')



import pandas as pd
import argparse


parser = argparse.ArgumentParser(description='Split .bam file to .sam files by mapped reads strand orientation')
parser.add_argument('file_in', metavar = 'in', type=str)
parser.add_argument('--prefix')

args = parser.parse_args()

prefix = args.prefix

read_type_num = {'INTERGENIC':0,    'INTRONIC':0,    'CODING':0,    'UTR':0,    'AMB':0}

strand_type_num = {
    'minus_minus': 0,
    'minus_plus': 0,
    'plus_plus': 0,
    'plus_minus': 0,
    'plus_AMB': 0,
    'minus_AMB': 0
}

out_file_names = {x: prefix + x + '.sam' for x in strand_type_num.keys()}

out_files = {x: open(out_file_names[x], 'w') for x in out_file_names.keys()}

def return_collapsed(it):
    # set has exactly 1 element, meaning that all elements are the same in the list 
    if len(set(it)) == 1:
        return it[0]
    else:
        return 'AMB'

with open(args.file_in, 'r') as fi:
    for line in fi:
        # if line is header line
        if line.startswith('@'):
            for f in out_files.values():
                f.write(line)

            # go to next iteration
            continue

        line_stripped = line.strip()

        elements = line_stripped.split()

        last = elements[-1]

        read_overlaps_gene = False

        # set gene strand
        if last.startswith('gs'):
            # if last element begins with gs, this means that read overlaps a gene (fw or rv strands)
            read_overlaps_gene = True
            gene_strand = return_collapsed(last.split(':')[-1].split(','))
            
            if gene_strand == '-':
                gene_strand = 'minus'
            elif gene_strand == '+':
                gene_strand = 'plus'

        else:
            gene_strand = 'AMB'

        # set read strand
        if elements[1] == '0':
            read_strand = 'plus'
        else:
            read_strand = 'minus'

        # get read type
        if read_overlaps_gene:
            read_type = return_collapsed(elements[-3].split(':')[-1].split(','))
        else:
            # if read do not overlap a gene, it is clearly intergenic
            read_type = 'INTERGENIC'


        read_type_num[read_type] = read_type_num[read_type] + 1
        
        strand_type = read_strand + '_' + gene_strand

        strand_type_num[strand_type] = strand_type_num[strand_type] + 1

        # print the read to the correct split file, depending on strand orientation
        out_files[strand_type].write(line)
        
with open(prefix + 'read_type_num.txt', 'w') as fo:
    for key, value in read_type_num.items():
        fo.write('%s %s\n' % (key, value))

with open(prefix + 'strand_type_num.txt', 'w') as fo:
    for key, value in strand_type_num.items():
        fo.write('%s %s\n' % (key, value))

for f in out_files.values():
    f.close()



import os
import logging
from more_itertools import grouper
from collections import OrderedDict, defaultdict
import numpy as np
from spacemake.util import rev_comp, fasta_chunks, ensure_path, read_fq


logger = logging.getLogger("spacemake.longread.util")


def load_oligos(fname=""):
    if not fname:
        base = os.path.dirname(__file__)
        fname = os.path.join(base, "../data/oligo_blocks.fa")

    blocks = OrderedDict()
    for fa_id, seq in fasta_chunks(open(fname)):
        blocks[fa_id] = seq
        blocks[fa_id + "_RC"] = rev_comp(seq)

    logger.info(f"load_oligos(): loaded {len(blocks)} sequences from '{fname}'")
    return blocks


def count_dict_collapse_misc(
    counts, misc_thresh=0.01, total=1, add_up=None, sig_intact=None
):
    out_counts = defaultdict(int)
    out_frac = defaultdict(float)

    misc = 0
    sum = 0
    if sig_intact is not None:
        complete = ",".join(sig_intact)
        everything = set(sig_intact)
    else:
        complete = None
        everything = set()

    def relkey(key):
        if sig_intact is None:
            return key

        if key == complete:
            return "complete"

        obs = set(key.split(","))
        there = obs & everything
        extra = obs - everything
        missing = everything - obs

        if len(missing) <= len(there):
            res = "missing_" + ",".join(sorted(missing))
        else:
            res = "only_" + ",".join(sorted(there))
        if extra:
            res += "_extra_" + ",".join(sorted(extra))

        return res

    for key, n in sorted(counts.items()):
        key = relkey(key)
        sum += n
        f = n / float(total)
        if f < misc_thresh:
            misc += n
        else:
            out_counts[key] = n
            out_frac[key] = f

    if misc > 0:
        out_counts["misc"] += misc
        out_frac["misc"] += misc / float(total)

    if add_up is None:
        other = total - sum
    else:
        other = total - counts[add_up]

    if other > 0:
        out_counts["N/A"] += other
        out_frac["N/A"] += other / float(total)

    return out_counts, out_frac


def count_dict_out(counts, title, misc_thresh=0.01, total=1, **kw):
    print(f"### {title}")
    colname = title.replace(" ", "_")
    out_counts, out_frac = count_dict_collapse_misc(counts, misc_thresh, total, **kw)
    for key, count in sorted(out_counts.items(), key=lambda x: -x[1]):
        print(f"{colname}\t{key}\t{count}\t{out_frac[key]:.3f}")

    return out_counts, out_frac


def count_dict_split(counts, pattern, name):
    import re

    out_d = defaultdict(int)
    for key, value in counts.items():
        if re.search(pattern, key):
            out_d[name] += value
        else:
            out_d[key] += value

    return out_d


def count_dict_to_df(counts, kind="", n_total=0):
    keys_values = sorted(counts.items(), key=lambda x: -x[1])
    import pandas as pd

    df = pd.DataFrame(data=keys_values, columns=["name", "count"])
    if n_total:
        df = df.append({"name": "n_total", "count": n_total}, ignore_index=True)
        df["fraction"] = df["count"] / n_total
    if kind:
        df["kind"] = kind
    return df


def count_dict_from_df(df, kind):
    df = df.query(f"kind == '{kind}'")
    # print(df)
    keys = df["name"]
    values = df["count"]
    return dict(zip(keys, values))


def process_intact_signature(complete_signature, prefixes=["P5"], suffixes=["N70X"]):
    complete = complete_signature.split(",")
    while complete[0] in prefixes:
        complete.pop(0)

    while complete[-1] in suffixes:
        complete.pop()

    complete_order = dict(x[::-1] for x in enumerate(complete))
    # print(f"complete={complete}")

    return tuple(complete), complete_order


def digest_signatures(
    sig_counts,
    bead_related="bead_start",
    complete_signature="P5,bead_start,OP1,polyT,N70X",
    prefixes=[
        "P5",
    ],
    suffixes=[
        "N70X",
    ],
):
    bead_counts = defaultdict(int)
    ov_counts = defaultdict(int)
    n_bead_related = 0

    complete, complete_order = process_intact_signature(
        complete_signature, prefixes, suffixes
    )
    complete_set = set(complete)
    found_part_counts = defaultdict(int)

    def describe(found_set):
        missing = complete_set - found_set
        if not missing:
            descr = "complete"
        elif len(missing) < len(found_set):
            descr = f"missing_{','.join(sorted(missing))}"
        else:
            descr = f"only_{','.join(sorted(found_set))}"

        return descr

    def bead_relation(parts):
        search = list(complete)
        at = 0

        try:
            i = parts.index(search[0])  # look for first part, e.g. bead_start
        except ValueError:
            i = 0

        found = []
        for part in parts[i:]:
            # find co-linear matches,
            # ignore extra inserted segments
            # (for now)
            if part in search[at:]:
                found.append(part)
                at = search.index(part)

        found_set = set(found)
        found_tup = tuple(sorted(found_set, key=lambda x: complete_order[x]))

        return describe(found_set), found_tup

    for sig, count in sig_counts.items():
        parts = sig.split(",")
        if bead_related in parts:
            br, found_tup = bead_relation(parts)
            bead_counts[br] += count
            n_bead_related += count

            for i in range(1, len(found_tup) + 1):
                found_part_counts[found_tup[:i]] += count
        else:
            ov_counts[sig] = count

    ov_counts["bead-related"] = n_bead_related
    return ov_counts, bead_counts, found_part_counts, complete


def gather_data_from_overview(df, samples, attr, na=np.nan):
    # print(df)
    # maybe not the most elegant, but doesn't crash on missing values.
    # TODO: look at pivot for a cleaner implementation
    x = np.arange(len(samples))
    # print(attr, samples)
    raw = [df.query(f"sample == '{s}'")[attr].values for s in samples]
    y = [r[0] if len(r) else na for r in raw]

    return x, y



__version__ = "0.9.6"
__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

import os
import sys
import logging
import pandas as pd
import numpy as np
import pysam
import argparse
from collections import defaultdict

import spacemake.reporting as rep

# TODO:
# * sampling of reads
# * graphical output of most common soft-clipped sequences
#   (MUSCLE?)


def gf_prio(gf, hierarchy=["CODING", "UTR", "INTRONIC", "INTERGENIC"]):
    parts = gf.split(",")
    if len(parts) == 1:
        # fast path if only one gf is given
        return gf

    parts = set(parts)
    for h in hierarchy:
        if h in parts:
            return h

    # could not resolve??
    return gf


def parse_args():
    parser = argparse.ArgumentParser("alnstats")
    parser.add_argument("fname", help="a gene-annotated BAM file")
    parser.add_argument(
        "--n-max",
        type=int,
        default=0,
        help="number of BAM records to process (default=0 -> all)",
    )
    parser.add_argument(
        "--out-csv", default="./", help="where to store the CSV raw stats"
    )
    parser.add_argument(
        "--out-pdf", default="./", help="where to store the PDF reports"
    )
    parser.add_argument(
        "--out-png", default="./", help="where to store the PNG reports"
    )
    args = parser.parse_args()

    return args


class Results:
    pass


def coarsegrain_CIGAR(cigar):
    coarse = cigar.replace("MNM", "M").replace("MDM", "M").replace("MIM", "M")
    if coarse == cigar:
        return cigar
    else:
        return coarsegrain_CIGAR(coarse)


def scan_bam(fname, n_max=0):
    res = Results()
    res.fname = fname
    res.n_max = n_max
    res.aln_types = defaultdict(int)
    res.cigar_types = defaultdict(int)
    res.tag_types = defaultdict(int)
    res.match_len_by_cigtype = defaultdict(lambda: defaultdict(int))
    res.match_len_by_tag = defaultdict(lambda: defaultdict(int))
    res.sc_lengths = defaultdict(lambda: defaultdict(int))
    res.sc_seq = defaultdict(lambda: defaultdict(int))

    op_dict = list("MIDNSHP=XB")
    sam = pysam.Samfile(fname, "rb")
    for r in sam.fetch(until_eof=True):
        res.aln_types["N_reads"] += 1
        if r.is_unmapped:
            res.aln_types["unmapped"] += 1
            continue

        if r.mapping_quality < 255:
            res.aln_types["multimapper"] += 1
        else:
            res.aln_types["unique"] += 1

        minus_strand = r.is_reverse
        if r.has_tag("gf"):
            tag = gf_prio(r.get_tag("gf"))
        else:
            tag = "INTERGENIC"

        res.tag_types[tag] += 1

        cigar = list(r.cigartuples)
        cigtype = "".join([op_dict[c[0]] for c in cigar])
        cigtype = coarsegrain_CIGAR(cigtype)  # ignore splicing
        res.cigar_types[cigtype] += 1

        if cigar[0][0] == pysam.CSOFT_CLIP:
            n = cigar[0][1]
            seq = r.query[:n]
            if minus_strand:
                res.sc_seq["3'"][seq] += 1
                res.sc_lengths["3'"][n] += 1
            else:
                res.sc_seq["5'"][seq] += 1
                res.sc_lengths["5'"][n] += 1

        if cigar[-1][0] == pysam.CSOFT_CLIP:
            n = cigar[-1][1]
            seq = r.query[-n:]
            if minus_strand:
                res.sc_seq["5'"][seq] += 1
                res.sc_lengths["5'"][n] += 1
            else:
                res.sc_seq["3'"][seq] += 1
                res.sc_lengths["3'"][n] += 1

        n_match = 0
        for op, n in r.cigartuples:
            if op == pysam.CMATCH:
                n_match += n

        res.match_len_by_cigtype[cigtype][n_match] += 1
        res.match_len_by_cigtype["all"][n_match] += 1
        res.match_len_by_tag[tag][n_match] += 1
        res.match_len_by_tag["all"][n_match] += 1

        if (n_max > 0) and (res.aln_types["N_reads"] >= n_max):
            break

    res.N_total = res.aln_types["N_reads"]

    return res


def make_plots(res):
    import matplotlib.pyplot as plt

    # preparing the plot
    fig, axes = plt.subplots(
        3, 3, figsize=(14, 11)
    )  # gridspec_kw={'height_ratios': [1, 1, 1]})
    axes = np.array(axes).flatten()

    # preparing count dictionaries for the donut plots (pie charts)
    cig_types_d, _ = rep.count_dict_collapse_misc(res.cigar_types, total=res.N_total)
    tag_types_d, _ = rep.count_dict_collapse_misc(res.tag_types, total=res.N_total)
    aln_types_d, _ = rep.count_dict_collapse_misc(
        res.aln_types, total=res.N_total, add_up="N_reads"
    )

    # First row: donut plots
    ciglabels, cigcolors = rep.donut_plot(
        axes[0], cig_types_d, title="CIGAR types", cmap="Accent"
    )

    taglabels, tagcolors = rep.donut_plot(
        axes[1], tag_types_d, title="gf types", cmap="Dark2"
    )

    alnlabels, alncolors = rep.donut_plot(
        axes[2],
        aln_types_d,
        title="Alignment status",
        labels=["unique", "multimapper", "unmapped"],
        cmap="Set2",
    )

    # Second row: histograms of no. matching bases
    rep.len_plot(
        axes[3],
        res.match_len_by_cigtype,
        labels=ciglabels + ["all"],
        colors=cigcolors + ["k"],
        title="CIGAR types",
    )

    rep.len_plot(
        axes[4],
        res.match_len_by_tag,
        labels=taglabels + ["all"],
        colors=tagcolors + ["k"],
        title="gf (annotation) types",
    )

    rep.len_plot(
        axes[5],
        res.sc_lengths,
        title="read position",
        xlabel="soft-clipped bases",
        colors=["r", "b"],
    )

    # Third row: cumulative histograms of no. matching bases
    rep.len_plot(
        axes[6],
        res.match_len_by_cigtype,
        labels=ciglabels + ["all"],
        colors=cigcolors + ["k"],
        title="CIGAR types",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )

    rep.len_plot(
        axes[7],
        res.match_len_by_tag,
        labels=taglabels + ["all"],
        colors=tagcolors + ["k"],
        title="gf (annotation) types",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )

    rep.len_plot(
        axes[8],
        res.sc_lengths,
        title="read position",
        colors=["r", "b"],
        xlabel="soft-clipped bases",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )
    return fig


def cmdline():
    args = parse_args()
    path = args.fname.split("/")

    if len(path) >= 4 and path[-4].startswith("sts"):
        sample_name = path[-4]
    else:
        sample_name = os.path.basename(args.fname).rsplit(".", maxsplit=1)[0]

    logging.info(f"Starting analysis of sample_name='{sample_name}'")
    res = scan_bam(args.fname, n_max=args.n_max)

    import matplotlib

    matplotlib.use("Agg")

    fig = make_plots(res)
    fig.suptitle(sample_name)
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.savefig(f"{args.out_pdf}/{sample_name}.pdf")
    fig.savefig(f"{args.out_png}/{sample_name}.png", dpi=300)

    # Done with the figure. Text output.
    # TODO: store as CSV instead

    rep.count_dict_out(
        res.aln_types, "alignment types", total=res.N_total, add_up="N_reads"
    )
    rep.count_dict_out(res.cigar_types, "CIGAR types", total=res.N_total)
    rep.count_dict_out(res.tag_types, "annotation (gf) types", total=res.N_total)

    # report the most frequent soft-clipped sequences
    def most_frequent(data, total, n=20, title="top"):
        print(f"### {title}")
        for k, v in sorted(data.items(), key=lambda x: -x[1])[:n]:
            # print(k, v, total)
            print(f"{k}\t{v}\t{100.0 * v / total:.3f} %")

    most_frequent(
        res.sc_seq["5'"],
        np.array(list(res.sc_seq["5'"].values())).sum(),
        title="most frequent 5' soft-clipped sequences",
    )
    most_frequent(
        res.sc_seq["3'"],
        np.array(list(res.sc_seq["3'"].values())).sum(),
        title="most frequent 3' soft-clipped sequences",
    )


if __name__ == "__main__":
    cmdline()



def calculate_adata_metrics(adata, dge_summary_path=None, n_reads=None):
    import scanpy as sc
    import pandas as pd
    # calculate mitochondrial gene percentage
    adata.var['mt'] = adata.var_names.str.startswith('Mt-') |\
            adata.var_names.str.startswith('mt-') |\
            adata.var_names.str.startswith('MT-')

    sc.pp.calculate_qc_metrics(adata, qc_vars=['mt'], percent_top=None, log1p=False, inplace=True)
    
    add_reads = False
    if dge_summary_path is not None:
        dge_summary = pd.read_csv(dge_summary_path, skiprows=7, sep ='\t', index_col = 'cell_bc', names = ['cell_bc', 'n_reads', 'n_umi', 'n_genes'])

        adata.obs = pd.merge(adata.obs, dge_summary[['n_reads']], left_index=True, right_index=True)

        add_reads = True

    if n_reads is not None:
        adata.obs['n_reads'] = n_reads
        add_reads = True

    if add_reads:
        adata.obs['reads_per_counts'] = adata.obs.n_reads / adata.obs.total_counts

def calculate_shannon_entropy_scompression(adata):
    import math
    import itertools
    import numpy as np
    from collections import Counter

    def compute_shannon_entropy(barcode):
        prob, length = Counter(barcode), float(len(barcode))
        return -sum( count/length * math.log(count/length, 2) for count in prob.values())

    def compute_string_compression(barcode):
        compressed_barcode = ''.join(
                letter + str(len(list(group)))
                for letter, group in itertools.groupby(barcode))

        return len(compressed_barcode)

    bc = adata.obs.index.to_numpy()
    bc_len = len(bc[0])
    theoretical_barcodes = np.random.choice(['A','C', 'T', 'G'],
        size = (bc.shape[0], bc_len))

    adata.obs['exact_entropy'] = np.round(np.array(
        [compute_shannon_entropy(cell_bc) for cell_bc in bc]), 2)
    adata.obs['theoretical_entropy'] = np.round(np.array(
        [compute_shannon_entropy(cell_bc) for cell_bc in theoretical_barcodes]), 2)
    adata.obs['exact_compression'] = np.round(np.array(
        [compute_string_compression(cell_bc) for cell_bc in bc]), 2)
    adata.obs['theoretical_compression'] = np.round(np.array(
        [compute_string_compression(cell_bc) for cell_bc in theoretical_barcodes]), 2)

def dge_to_sparse_adata(dge_path, dge_summary_path):
    import anndata
    import numpy as np
    import gzip
    import pandas as pd
    from scipy.sparse import csr_matrix, vstack

    ix = 0
    mix = 0
    matrices = []
    gene_names = []

    with gzip.open(dge_path, 'rt') as dge:
        first_line = dge.readline().strip().split('\t')

        barcodes = first_line[1:]
        ncol = len(barcodes)

        # create an intermediate matrix, which will contain 1000 rows
        # and cell_number columns. we will parse the dge file and fill
        # this matrix line by line. Like this we create gene_number/1000 number
        # of CSR (compressed sparse row) matrices which we join at the end
        M = np.zeros((1000, ncol))

        # read DGE line by line
        # first row: contains CELL BARCODEs
        # each next row contains one gene name, and the counts of that gene
        for line in dge:
            vals = line.strip().split('\t')
            # first element contains the gene name
            # append gene name to gene_names
            gene_names.append(vals[0])
            vals = vals[1:]

            # store counts as np.array
            vals = np.array(vals, dtype=np.int64)

            # update the 1000xcell_number matrix 
            M[ix] = vals

            if ix % 1000 == 999:
                # if we reached the end of M, make it sparse and append to other
                # already read matrices
                mix = mix + 1
                matrices.append(csr_matrix(M))
                ix = 0

                # reset M
                M = np.zeros((1000, ncol))
            else:
                ix = ix + 1

        # get the leftovers: these are the overhang lines, when gene_number is
        # not divisible by 1000
        M = M[:ix]
        matrices.append(csr_matrix(M))

        # sparse expression matrix
        X = vstack(matrices, format='csr')
        
        # create anndata object, but we get the transpose of X, so matrix will
        # be in CSC format
        adata = anndata.AnnData(X.T,
                obs = pd.DataFrame(index=barcodes),
                var = pd.DataFrame(index=gene_names))

        # name the index
        adata.obs.index.name = 'cell_bc'

        # attach metrics such as: total_counts, pct_mt_counts, etc
        # also attach n_genes, and calculate pcr
        calculate_adata_metrics(adata, dge_summary_path)

        # calculate per shannon_entropy and string_compression per bead
        calculate_shannon_entropy_scompression(adata)

        return adata

def load_external_dge(dge_path):
    import anndata
    import scanpy as sc

    from scanpy._utils import check_nonnegative_integers
    from scipy.sparse import issparse, csc_matrix
    from spacemake.errors import SpacemakeError

    adata = sc.read(dge_path)

    if not check_nonnegative_integers(adata.X):
        raise SpacemakeError(f'External dge seems to contain values '+
            'which are already normalised. Raw-count matrix expected.')

    if not issparse(adata.X):
        adata.X = csc_matrix(adata.X)

    # name the index
    adata.obs.index.name = 'cell_bc'

    # attach metrics such as: total_counts, pct_mt_counts, etc
    # also attach n_genes, and calculate pcr
    calculate_adata_metrics(adata)

    return adata

def parse_barcode_file(barcode_file):
    import pandas as pd

    bc = pd.read_csv(barcode_file, sep='[,|\t]', engine='python')

    # rename columns
    bc = bc.rename(columns={'xcoord':'x_pos', 'ycoord':'y_pos','barcodes':'cell_bc', 'barcode':'cell_bc'})\
        .set_index('cell_bc')\
        .loc[:, ['x_pos', 'y_pos']]

    bc = bc.loc[~bc.index.duplicated(keep='first')]

    bc = bc.loc[~bc.index.duplicated(keep='first')]

    return bc

def attach_barcode_file(adata, barcode_file):
    import pandas as pd

    bc = parse_barcode_file(barcode_file)

    # new obs has only the indices of the exact barcode matches
    new_obs = adata.obs.merge(bc, left_index=True, right_index=True, how='inner')
    adata = adata[new_obs.index, :]
    adata.obs = new_obs
    adata.obsm['spatial'] = adata.obs[['x_pos', 'y_pos']].to_numpy()

    return adata




import logging
import novosparc

from anndata import AnnData
from spacemake.util import message_aggregation
from numpy import ndarray

logger_name = 'spacemake.spatial.novosparc_reconstruction'
logger = logging.getLogger(logger_name)

def get_parser():
    import argparse

    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description='reconstruct 2-d tissue with novosparc',
    )

    parser.add_argument(
        '--single_cell_dataset', type=str, help='path to single_cell.h5ad file',
        required = True,
    )

    parser.add_argument(
        '--spatial_dataset', type=str, help='path to spatial.h5ad file',
        required=False,
    )

    parser.add_argument(
        '--output', type=str, help='path to output.h5ad file',
        required=True,
    )

    return parser

# copy of a yet-to-be-pushed novosparc function
def quantify_clusters_spatially(tissue: novosparc.cm.Tissue, cluster_key: str='clusters') -> ndarray:
    """Maps the annotated clusters obtained from the scRNA-seq analysis onto
    the tissue space.

    :param tissue:    the novosparc tissue object containing the gene expression data,
                      the clusters annotation and the spatial reconstruction. Assumes
                      that the cluster annotation exists in the underlying anndata object.
    :type tissue: novosparc.cm.Tissue
    :param cluster_key: the key of the clustering to be used for deduction
    :type cluster_key: str
    rtype: numpy.ndarray
    """
    import numpy as np

    clusters = tissue.dataset.obs[cluster_key].to_numpy().flatten()
    cluster_names = np.unique(clusters)
    ixs = np.array(
            [np.argmax(
                np.array(
                    [np.median(
                        np.array(
                            tissue.gw[:, location][np.argwhere(clusters == cluster).flatten()]))
                                         for cluster in cluster_names]))
                    for location in range(len(tissue.locations))])
    return [cluster_names[ix] for ix in ixs]
    
def novosparc_denovo(
        adata: AnnData,
        num_spatial_locations: int=5000,
        num_input_cells: int=30000,
        locations=None,
    ) -> novosparc.cm.Tissue:
    """reconstruct spatial information de-novo with novosparc.

    :param adata: A spacemake processed sample.
    :type adata: AnnData
    :param num_spatial_locations:
    :type num_spatial_locations: int
    :param num_input_cells:
    :type num_input_cells: int
    :param locations: Numpy array of (x,y) locations (optional)
    :type locations: numpy.ndarray
    :rtype: novosparc.cm.Tissue
    """
    import numpy as np
    import pandas as pd
    import scanpy as sc

    import novosparc
    import anndata

    from scipy.sparse import issparse, csc_matrix

    logger.info('Reconstructing the tissue de-novo with novosparc') 
    
    gene_names = adata.var.index.tolist()

    num_cells, num_genes = adata.shape

    if num_cells > num_input_cells:
        sc.pp.subsample(adata, n_obs = num_input_cells)

    if num_cells < num_spatial_locations:
        num_spatial_locations = num_cells

    sc.pp.highly_variable_genes(adata, n_top_genes = 100)
    is_var_gene = adata.var['highly_variable']
    # select only 100 genes
    var_genes = list(is_var_gene.index[is_var_gene])

    dge_rep = adata.to_df()[var_genes]

    # if you uncomment this, adata.X will be stored not as sparse matrix, rather
    # as a regular numpy array. uncommenting this doesnt throw an error
    if issparse(adata.X):
        dense_adata = anndata.AnnData(
            adata.X.toarray(),
            obs = adata.obs,
            var = adata.var)
        adata = dense_adata
        del dense_adata

    if locations is None:
        # create circle locations
        locations = novosparc.gm.construct_circle(num_locations = num_spatial_locations)

    tissue = novosparc.cm.Tissue(dataset=adata, locations=locations)

    num_neighbors_s = num_neighbors_t = 5

    logger.info('Novosparc setup')
    tissue.setup_smooth_costs(dge_rep=dge_rep, num_neighbors_s=num_neighbors_s, num_neighbors_t=num_neighbors_t)

    tissue.reconstruct(alpha_linear=0, epsilon=5e-3)

    return tissue

def novosparc_mapping(sc_adata: AnnData, st_adata: AnnData) -> novosparc.cm.Tissue:
    """novosparc_mapping.

    :param sc_adata:
    :type sc_adata: AnnData
    :param st_adata:
    :type st_adata: AnnData
    :rtype: novosparc.cm.Tissue
    """
    import scanpy as sc
    import anndata
    import novosparc

    from scanpy._utils import check_nonnegative_integers
    from scipy.sparse import csc_matrix

    logger.info('Mapping single-cell data onto spatial data with novosparc')

    if (check_nonnegative_integers(sc_adata.X)
        or check_nonnegative_integers(st_adata.X)
    ):
        # if any of the inputs is count-data, raise error
        raise SpacemakeError(f'External dge seems to contain raw counts. '+
            'Normalised values are expected for both sc_adata and st_adata.')

    # calculate top 500 variable genes for both
    sc.pp.highly_variable_genes(sc_adata, n_top_genes=500)
    sc.pp.highly_variable_genes(st_adata, n_top_genes=500)

    sc_adata_hv = sc_adata.var_names[sc_adata.var.highly_variable].to_list()
    st_adata_hv = st_adata.var_names[st_adata.var.highly_variable].to_list()

    markers = list(set(sc_adata_hv).intersection(st_adata_hv))

    logger.info(f'{len(markers)} number of common markers found. Using them' +
        ' for reconstruction')

    # save sc dge as a pandas dataframe
    dge_rep = sc_adata.to_df()[sc_adata_hv]

    if not 'spatial' in st_adata.obsm:
        raise SpacemakeError(f'The object provided to st_adata is not spatial')

    locations = st_adata.obsm['spatial']
    atlas_matrix = st_adata.to_df()[markers].values

    # make dense dataset
    dense_dataset = anndata.AnnData(
        sc_adata.X.toarray(),
        obs = sc_adata.obs,
        var = sc_adata.var)

    marker_ix = [dense_dataset.var.index.get_loc(marker) for marker in markers]

    tissue = novosparc.cm.Tissue(dataset=dense_dataset, locations=locations)
    num_neighbors_s = num_neighbors_t = 5

    tissue.setup_linear_cost(markers_to_use=marker_ix, atlas_matrix=atlas_matrix,
                             markers_metric='minkowski', markers_metric_p=2)
    tissue.setup_smooth_costs(dge_rep = dge_rep,
                              num_neighbors_s=num_neighbors_s,
                              num_neighbors_t=num_neighbors_t)

    tissue.reconstruct(alpha_linear=0.5, epsilon=5e-3)

    return tissue

def save_novosparc_res(tissue, adata_original) -> AnnData:
    import anndata
    import pandas as pd
    import numpy as np

    from scipy.sparse import csc_matrix

    adata_reconst = anndata.AnnData(
        csc_matrix(tissue.sdge.T),
        var = pd.DataFrame(index=tissue.dataset.var_names))

    logger.info('Scaling mapped data to original data...')

    adata_reconst.X = np.sum(adata_original.X) * adata_reconst.X / np.sum(adata_reconst.X)

    adata_reconst.obsm['spatial'] = tissue.locations

    logger.info('Transferring original cluster labels...')

    for res_key in adata_original.obs.columns[adata_original.obs.columns.str.startswith('leiden_')]:
        adata_reconst.obs[f'novosparc_{res_key}'] = quantify_clusters_spatially(tissue, res_key)

    return adata_reconst

@message_aggregation(logger_name)
def cmdline():
    import scanpy as sc

    parser = get_parser()

    args = parser.parse_args()

    sc_adata = sc.read(args.single_cell_dataset)

    if args.spatial_dataset is not None:
        # make mapping
        st_adata = sc.read(args.spatial_dataset)

        tissue_reconst = novosparc_mapping(sc_adata, st_adata)
        adata = save_novosparc_res(tissue_reconst, sc_adata)
    else:
        tissue_reconst = novosparc_denovo(sc_adata)  
        adata = save_novosparc_res(tissue_reconst, sc_adata)

    adata.write(args.output)

if __name__ == "__main__":
    cmdline()



################################
# Final output file generation #
################################
def get_output_files(
    pattern,
    projects=[],
    samples=[],
    filter_merged=False,
    run_on_external=True,
    **kwargs,
):
    out_files = []
    df = project_df.df

    if projects != [] or samples != []:
        project_df.assert_index_value(projects, "project_id")

        project_df.assert_index_value(samples, "sample_id")

        ix = project_df.get_ix_from_project_sample_list(
            project_id_list=projects, sample_id_list=samples
        )

        df = df.loc[ix]

    if filter_merged:
        df = df.loc[~df.is_merged]

    for index, row in df.iterrows():
        is_external = project_df.is_external(
            project_id=index[0], sample_id=index[1]
        )

        has_dge = project_df.has_dge(
            project_id=index[0], sample_id=index[1]
        )

        if not has_dge:
            # if there is no dge, skip
            continue
        
        if not run_on_external and is_external:
            # if we do not want to run this on external samples
            # and sample is external, skip
            continue

        for run_mode in row["run_mode"]:
            run_mode_variables = project_df.config.get_run_mode(run_mode).variables

            out_files = out_files + expand(
                pattern,
                project_id=index[0],
                sample_id=index[1],
                puck_id=row["puck_id"],
                run_mode=run_mode,
                umi_cutoff=run_mode_variables["umi_cutoff"],
                **kwargs,
            )

    return out_files

def get_all_dges(wildcards):
    df = project_df.df

    dges = []

    for index, row in df.iterrows():
        for run_mode in row["run_mode"]:
            if project_df.has_dge(project_id=index[0], sample_id=index[1]):
                dges.append(
                    get_dge_from_run_mode(
                        project_id=index[0],
                        sample_id=index[1],
                        run_mode=run_mode,
                        data_root_type="complete_data",
                        downsampling_percentage="",
                    )["dge"],
                )

    return dges


def get_raw_dge(wildcards):
    is_external = project_df.is_external(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    has_dge = project_df.has_dge(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    if has_dge:
        out_files = {}
    else:
        return []

    if is_external:
        out_files["dge"] = project_df.get_metadata(
            "dge", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        out_files["dge"] = dge_out
        out_files["dge_summary"] = dge_out_summary

    return out_files


def get_reads(wildcards):
    ###
    # R1 and R2 for demultiplexed reads will return none
    ###
    reads = project_df.get_metadata(
        "R" + wildcards.mate, sample_id=wildcards.sample_id, project_id=wildcards.project_id
    )
    if reads is None or reads == []:
        return ["none"]
    else:
        # reads already
        return reads


# barcode flavor parsing and query functions
class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def parse_barcode_flavors(
    config,
    bc_default_settings=dict(
        bc1_ref="",
        bc2_ref="",
        cell_raw="None",
        score_threshold=0.0,
        min_opseq_score=22,
        bam_tags="CR:{cell},MI:{UMI}",
    ),
):
    """
    Reads the 'barcode_flavor' from 'knowledge' section of the config.yaml.
    parses and gathers the settings for barcode flavors
    """
    preprocess_settings = {}
    for flavor, flavor_settings in config["barcode_flavors"].items():
        # for each flavor, also retrieve the configuration
        # first make a copy of the default values
        d = dict(bc_default_settings)
        d.update(flavor_settings)
        preprocess_settings[flavor] = dotdict(d)

    res = dotdict(
        dict(
            preprocess_settings=preprocess_settings,
        )
    )

    return res


# all barcode flavor info from config.yaml
# is kept here for convenient lookup
bc_flavor_data = parse_barcode_flavors(config)


def get_bc_preprocess_settings(wildcards):
    """
    This function will return a dictionary of information
    on the read1 preprocessing, according to barcode_flavor
    """
    flavor = project_df.get_metadata(
        "barcode_flavor", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )
    if flavor not in bc_flavor_data.preprocess_settings:
        raise Exception(flavor)

    settings = bc_flavor_data.preprocess_settings[flavor]

    return settings


def get_demux_indicator(wildcards):
    demux_dir = project_df.get_metadata(
        "demux_dir", sample_id=wildcards.sample_id, project_id=wildcards.project_id
    )

    return expand(demux_indicator, demux_dir=demux_dir)


def get_star_input_bam(wildcards):
    if wildcards.polyA_adapter_trimmed == ".polyA_adapter_trimmed":
        return {"reads": tagged_polyA_adapter_trimmed_bam}
    else:
        return {"reads": tagged_bam}


def get_final_bam(wildcards):
    is_merged = project_df.get_metadata(
        "is_merged", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    if is_merged:
        return [final_merged_bam]
    else:
        return [final_bam]


def get_dge_input_bam(wildcards):
    if wildcards.data_root_type == "complete_data":
        final_bam_pipe = final_bam_mm_included_pipe
        final_bam = get_final_bam(wildcards)
    elif wildcards.data_root_type == "downsampled_data":
        final_bam_pipe = downsampled_bam_mm_included_pipe
        final_bam = downsampled_bam

    if wildcards.mm_included == ".mm_included":
        out = {"reads": final_bam_pipe}
    else:
        out = {"reads": final_bam}

    return out


def get_species_genome_annotation(wildcards):
    # This function will return 2 things required by STAR:
    #    - annotation (.gtf file)
    #    - genome (.fa file)
    if "species" not in wildcards.keys():
        species = project_df.get_metadata(
            "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        species = wildcards.species

    files = project_df.config.get_variable("species", name=species)

    return {'genome': files['genome'],
            'annotation': files['annotation']}


def get_star_index(wildcards):
    # This function will return 1 things required by STAR:
    #    - index directory
    species = project_df.get_metadata(
        "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )
    species_data = project_df.config.get_variable("species", name=species)

    if 'STAR_index_dir' in species_data:
        return {'index': species_data['STAR_index_dir']}
    else:
        return {"index": expand(star_index, species=species)[0]}


def get_rRNA_genome(wildcards):
    files = project_df.config.get_variable("species", name=wildcards.species)

    return [files["rRNA_genome"]]


def get_bt2_rRNA_index(wildcards):
    species = project_df.get_metadata(
        "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    files = project_df.config.get_variable("species", name=species)

    if "rRNA_genome" in files:
        return {"index": expand(bt2_rRNA_index_dir, species=species)[0]}

    return []


def get_run_modes_from_sample(project_id, sample_id):
    run_mode_names = project_df.get_metadata(
        "run_mode", project_id=project_id, sample_id=sample_id
    )

    run_modes = {}

    for run_mode in run_mode_names:
        run_modes[run_mode] = project_df.config.get_run_mode(run_mode).variables

    return run_modes


def get_dge_extra_params(wildcards):
    dge_type = wildcards.dge_type

    extra_params = ""

    if dge_type == ".exon":
        extra_params = ""
    elif dge_type == ".intron":
        extra_params = "LOCUS_FUNCTION_LIST=null LOCUS_FUNCTION_LIST=INTRONIC"
    elif dge_type == ".all":
        extra_params = "LOCUS_FUNCTION_LIST=INTRONIC"
    if dge_type == ".Reads_exon":
        extra_params = "OUTPUT_READS_INSTEAD=true"
    elif dge_type == ".Reads_intron":
        extra_params = (
            "OUTPUT_READS_INSTEAD=true LOCUS_FUNCTION_LIST=null"
            + "LOCUS_FUNCTION_LIST=INTRONIC"
        )
    elif dge_type == ".Reads_all":
        extra_params = "OUTPUT_READS_INSTEAD=true LOCUS_FUNCTION_LIST=INTRONIC"

    if wildcards.mm_included == ".mm_included":
        extra_params = extra_params + " READ_MQ=0"

    return extra_params


def get_files_to_merge(pattern, project_id, sample_id, **kwargs):
    # recursive function to find all files to merge. a merged sample can be merged
    # from merged samples. to avoid cyclic dependencies, here we look for all files
    # which are the dependencies of the underlying samples
    is_merged = project_df.get_metadata(
        "is_merged", project_id=project_id, sample_id=sample_id
    )

    files = []

    if not is_merged:
        files = expand(pattern, sample_id=sample_id, project_id=project_id, **kwargs)
    else:
        merge_ix = project_df.get_metadata(
            "merged_from", sample_id=sample_id, project_id=project_id
        )

        for (p, s) in merge_ix:
            files = files + get_files_to_merge(
                project_id=p, sample_id=s, pattern=pattern, **kwargs
            )

    return list(set(files))


def get_files_to_merge_snakemake(pattern):
    # inner function to be returned
    def get_merged_pattern(wildcards):
        kwargs = {}

        # konvert wildcards to dict
        for key, value in wildcards.items():
            kwargs[key] = value

        files = get_files_to_merge(pattern=pattern, **kwargs)

        return files

    return get_merged_pattern


def get_ribo_depletion_log(wildcards):
    is_merged = project_df.get_metadata(
        "is_merged", sample_id=wildcards.sample_id, project_id=wildcards.project_id
    )

    if is_merged:
        return [merged_ribo_depletion_log]
    else:
        return [ribo_depletion_log]


def get_top_barcodes(wildcards):
    if wildcards.n_beads == "spatial":
        return {"top_barcodes": spatial_barcodes}
    if wildcards.dge_cleaned == "":
        return {"top_barcodes": top_barcodes}
    else:
        return {"top_barcodes": top_barcodes_clean}


def get_parsed_puck_file(wildcards):
    is_spatial = project_df.is_spatial(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    if is_spatial:
        return {"puck_file": parsed_spatial_barcodes}
    else:
        return []


def get_dge_from_run_mode(
    project_id, sample_id, run_mode, data_root_type, downsampling_percentage
):
    has_dge = project_df.has_dge(
        project_id=project_id, sample_id=sample_id)

    if not has_dge:
        raise SpacemakeError(
            f'Sample with id (project_id, sample_id)={project_id}, {sample_id})' +
            f' does not have a DGE')

    is_spatial = project_df.is_spatial(project_id=project_id, sample_id=sample_id)

    is_external = project_df.is_external(project_id=project_id, sample_id=sample_id)

    run_mode_variables = project_df.config.get_run_mode(run_mode).variables

    dge_type = ""
    dge_cleaned = ""
    polyA_adapter_trimmed = ""
    mm_included = ""

    # assign wildcards only for internal samples
    if not is_external:
        if run_mode_variables["polyA_adapter_trimming"]:
            polyA_adapter_trimmed = ".polyA_adapter_trimmed"

        if run_mode_variables["count_intronic_reads"]:
            dge_type = ".all"
        else:
            dge_type = ".exon"

        if run_mode_variables["count_mm_reads"]:
            mm_included = ".mm_included"

        if run_mode_variables["clean_dge"]:
            dge_cleaned = ".cleaned"

    if run_mode_variables["mesh_type"] == "hexagon":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = "hexagon"
    elif run_mode_variables["mesh_type"] == "circle":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = run_mode_variables["mesh_spot_distance_um"]

    external_wildcard = ""

    n_beads = run_mode_variables["n_beads"]

    if is_external:
        n_beads = "external"
        external_wildcard = ".external"

    if is_spatial:
        n_beads = "spatial"

    # select which pattern
    # if sample is not spatial, we simply select the normal, umi_filtered
    # dge, with the top_n barcodes
    # otherwise, if sample is spatial, either we return he whole dge, containing
    # all beads, or the a meshgrid
    if not is_spatial:
        dge_out_pattern = dge_out_h5ad
        dge_out_summary_pattern = dge_out_h5ad_obs
    elif run_mode_variables["mesh_data"]:
        dge_out_pattern = dge_spatial_mesh
        dge_out_summary_pattern = dge_spatial_mesh_obs
    else:
        dge_out_pattern = dge_spatial
        dge_out_summary_pattern = dge_spatial_obs

    out_files_pattern = {"dge_summary": dge_out_summary_pattern, "dge": dge_out_pattern}

    out_files = {
        key: expand(
            pattern,
            project_id=project_id,
            sample_id=sample_id,
            dge_type=dge_type,
            dge_cleaned=dge_cleaned,
            polyA_adapter_trimmed=polyA_adapter_trimmed,
            mm_included=mm_included,
            spot_diameter_um=spot_diameter_um,
            spot_distance_um=spot_distance_um,
            n_beads=n_beads,
            is_external=external_wildcard,
            data_root_type=data_root_type,
            downsampling_percentage=downsampling_percentage,
        )
        for key, pattern in out_files_pattern.items()
    }

    return out_files


def get_qc_sheet_input_files(wildcards):
    # returns star_log, reads_type_out, strand_info
    # first checks the run modes, and returns either polyA_adapter_trimmed, untrimmed
    # or both
    project_id = wildcards.project_id
    sample_id = wildcards.sample_id

    is_merged = project_df.get_metadata(
        "is_merged", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    run_modes = get_run_modes_from_sample(wildcards.project_id, wildcards.sample_id)

    is_polyA_adapter_trimmed = set(
        [x["polyA_adapter_trimming"] for x in run_modes.values()]
    )

    # if sample has both polyA trimmed and untrimmed mapped bam files
    if len(is_polyA_adapter_trimmed) == 2:
        polyA_adapter_trimmed_wildcard = ["", ".polyA_adapter_trimmed"]
    elif True in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [".polyA_adapter_trimmed"]
    elif False in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [""]

    extra_args = {
        "sample_id": wildcards.sample_id,
        "project_id": wildcards.project_id,
        "polyA_adapter_trimmed": polyA_adapter_trimmed_wildcard,
    }

    if is_merged:
        star_log_pattern = merged_star_log_file
    else:
        star_log_pattern = star_log_file

    to_return = {
        "star_log": expand(star_log_pattern, **extra_args),
        "reads_type_out": expand(reads_type_out, **extra_args),
        "strand_info": expand(strand_info, **extra_args),
    }

    for run_mode in run_modes:
        run_mode_dge = get_dge_from_run_mode(
            project_id,
            sample_id,
            run_mode,
            data_root_type=wildcards.data_root_type,
            downsampling_percentage=wildcards.downsampling_percentage,
        )

        to_return[f"{run_mode}.dge_summary"] = run_mode_dge["dge_summary"]

    return to_return


def get_bam_tag_names(project_id, sample_id):
    barcode_flavor = project_df.get_metadata(
        "barcode_flavor", project_id=project_id, sample_id=sample_id
    )

    bam_tags = config["barcode_flavors"][barcode_flavor]["bam_tags"]

    tag_names = {}

    for tag in bam_tags.split(","):
        tag_name, tag_variable = tag.split(":")

        tag_names[tag_variable] = tag_name

    return tag_names


def get_puck_file(wildcards):
    if not project_df.is_spatial(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    ):

        return []

    puck_barcode_file = project_df.get_metadata(
        "puck_barcode_file", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    puck_name = project_df.get_metadata(
        "puck", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )
    
    puck = project_df.config.get_puck(puck_name)

    if puck_barcode_file is None:
        return {"barcode_file": puck.variables['barcodes']}
    else:
        return {"barcode_file": puck_barcode_file}


def get_automated_analysis_dge_input(wildcards):
    # there are three options:
    # 1) no spatial dge
    # 2) spatial dge, no mesh
    # 3) spatial dge with a mesh
    return [
        get_dge_from_run_mode(
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            run_mode=wildcards.run_mode,
            data_root_type=wildcards.data_root_type,
            downsampling_percentage=wildcards.downsampling_percentage,
        )["dge"]
    ]

def get_novosparc_input_files(config):
    if ('reference_project_id' in config and
        config['reference_project_id'] != '' and
        'reference_sample_id' in config and
        config['reference_sample_id'] != '' and
        'reference_run_mode' in config and
        config['reference_run_mode'] != '' and
        'reference_umi_cutoff' in config and
        config['reference_umi_cutoff'] != ''):

        ret = expand(novosparc_with_reference_h5ad,
            data_root_type = 'complete_data',
            downsampling_percentage = '',
            project_id = config['project_id'],
            sample_id = config['sample_id'],
            run_mode = config['run_mode'],
            umi_cutoff = config['umi_cutoff'],
            reference_project_id = config['reference_project_id'],
            reference_sample_id = config['reference_sample_id'],
            reference_run_mode = config['reference_run_mode'],
            reference_umi_cutoff = config['reference_umi_cutoff'],
        )

    elif ('project_id' in config and
          'sample_id' in config and
          'umi_cutoff' in config and
          'run_mode' in config):
        ret = expand(novosparc_denovo_h5ad,
            data_root_type = 'complete_data',
            downsampling_percentage = '',
            project_id = config['project_id'],
            sample_id = config['sample_id'],
            run_mode = config['run_mode'],
            umi_cutoff = config['umi_cutoff'],
        )
    else:
        ret = []
    
    return ret

def get_novosparc_with_reference_input_files(wildcards):
    out_dict = {}

    st_adata = expand(automated_analysis_result_file,
        data_root_type = wildcards.data_root_type,
        downsampling_percentage = wildcards.downsampling_percentage,
        # access data without the first underscore
        project_id = wildcards.reference_project_id,
        sample_id = wildcards.reference_sample_id,
        run_mode = wildcards.reference_run_mode,
        umi_cutoff = wildcards.reference_umi_cutoff,
    )

    out_dict['st_adata'] = st_adata

    return out_dict




__author__ = ["Marvin Jens"]
__email__ = ["marvin.jens@mdc-berlin.de"]

import os
import argparse
import logging
import pandas as pd
import numpy as np
import spacemake.longread.util as util
import spacemake.longread.report as report
import spacemake.longread.cache as cache
import spacemake.longread.annotation as ann
from collections import defaultdict


def detect_sample(args):
    if args.sample is None:
        if not hasattr(args.fname):
            raise ValueError("please provide a sample-identifier via --sample")

        sample_name, _ = os.path.splitext(os.path.basename(args.fname))
        logging.info(f"auto-detected sample_name={sample_name}")
    else:
        sample_name = args.sample
    return sample_name.replace(".stats", "")


def store_results(df, path, fname, logger, **kw):
    fpath = os.path.join(util.ensure_path(path), fname)
    df.to_csv(fpath, sep="\t", **kw)
    logger.info(f"storing {len(df)} rows to '{fpath}'")
    return fpath


def load_results(path, fname, logger, **kw):
    fpath = os.path.join(path, fname)
    df = pd.read_csv(fpath, sep="\t", comment="#", **kw).fillna("other")
    logger.info(f"loaded {len(df)} rows from '{fpath}'")
    return df


class SignatureDB:
    def __init__(self, intact={}, related={}, color={}, label={}, prio={}):
        self.intact = intact
        self.related = related
        self.color = color
        self.label = label
        self.prio = prio

    @classmethod
    def from_YAML(cls, fname="samples.yaml"):
        import yaml

        logger = logging.getLogger("spacemake.longread.SampleDB.from_YAML")
        logger.info(f"reading longread signature definitions from '{fname}'")

        groups = yaml.load(open(fname), Loader=yaml.SafeLoader)
        grp = groups["signatures"]
        default = grp[groups["default"]]
        # print("default THING", default)

        intact_lkup = defaultdict(lambda: default["intact_bead"])
        brelated_lkup = defaultdict(lambda: default["bead_related"])
        color_lkup = defaultdict(lambda: default["color"])
        label_lkup = defaultdict(lambda: default["label"])
        prio_lkup = defaultdict(lambda: default["prio"])
        for name, d in groups["signatures"].items():
            # print(f"name={name} d={d}")
            intact_lkup[name] = d["intact_bead"]
            brelated_lkup[name] = d["bead_related"]
            color_lkup[name] = d["color"]
            label_lkup[name] = d["label"]
            prio_lkup[name] = d["prio"]

        # print(color_lkup)
        # print("DEFAULT COLOR", color_lkup["TEST"])
        logger.info(
            f"found {len(intact_lkup)} signature definitions: {sorted(intact_lkup.keys())}."
        )
        return cls(intact_lkup, brelated_lkup, color_lkup, label_lkup, prio_lkup)

    def sort_samples(self, samples, signatures):
        return sorted(
            zip(samples, signatures), key=lambda x: (self.prio.get(x[1], np.inf), x[0])
        )


def get_signature_db(args):
    if os.access(args.config, os.R_OK):
        cfg = args.config
    else:
        cfg = os.path.join(os.path.dirname(__file__), "../data/config/longread.yaml")

    return SignatureDB.from_YAML(cfg)


def setup_namespace(args):
    "perform initialization code common to more than one sub-command"
    d = vars(args)
    sample = detect_sample(args)
    d["sample_name"] = sample

    db = get_signature_db(args)
    d["signature_db"] = db

    intact_bead = db.intact[args.signature]
    d["intact_bead"] = intact_bead
    d["sig_intact"] = intact_bead.split(",")
    d["sig_core"], d["sig_core_order"] = util.process_intact_signature(intact_bead)

    d["bead_related"] = db.related[args.signature]
    # bead_related = args.bead_related
    # if bead_related is None:
    #     sig_intact = tuple(args.intact_bead.split(","))
    #     bead_related = sig_intact[0]

    d["logger"] = logging.getLogger(f"spacemake.longread.cmd.{args.subcmd}")
    d["blocks"] = util.load_oligos(args.blocks)

    from types import SimpleNamespace

    return SimpleNamespace(**d)


def aln_main(args):
    args = setup_namespace(args)

    cache.fill_caches(
        args.fname,
        args.sample_name,
        args.blocks,
        path=util.ensure_path(args.cache),
        n_proc=args.parallel,
    )
    df = cache.annotate(
        args.fname, args.sample_name, args.blocks, path=util.ensure_path(args.cache)
    )
    df.to_csv(
        os.path.join(
            util.ensure_path(args.annotation_out), f"{args.sample_name}.annotation.tsv"
        ),
        sep="\t",
        index=False,
    )


def ann_main(args):
    args = setup_namespace(args)
    args.logger.info(
        f"analyzing {args.sample_name} for intact signature {args.sig_intact}"
    )
    annotation = ann.AnnotatedSequences(
        args.fname,
        os.path.join(args.annotation_out, f"{args.sample_name}.annotation.tsv"),
        args.sample_name,
        args.blocks,
        min_score=args.min_score,
        orient_by=args.bead_related,
    )
    n_total = len(annotation.raw_sequences)
    logging.info(
        f"total number of reads in {args.fname} ({args.sample_name}) is {n_total}"
    )

    sig_counts, n_concat, n_reprimed = annotation.count_signatures()
    util.count_dict_out(sig_counts, "common signatures", total=n_total)
    print(
        f"n_concat={n_concat} ({100.0 * n_concat/n_total:.2f}%) "
        f"n_reprimed={n_reprimed} ({100.0 * n_reprimed/n_total:.2f}%)"
    )
    df_sig = util.count_dict_to_df(sig_counts, "signatures", n_total=n_total)

    concat_counts, n_occurrences = annotation.count_concatenations()
    util.count_dict_out(concat_counts, "concatenations", total=n_occurrences)
    df_concat = util.count_dict_to_df(
        concat_counts, "concatenations", n_total=n_occurrences
    )

    reprimed_counts = annotation.count_repriming()
    util.count_dict_out(reprimed_counts, "repriming", total=n_total)
    df_reprimed = util.count_dict_to_df(reprimed_counts, "repriming", n_total=n_total)

    partial_counts, prefixes, suffixes, pT_counts = annotation.completeness(
        args.sig_core, polyT=args.polyT
    )

    partial_counts_simple, _ = util.count_dict_collapse_misc(
        partial_counts, sig_intact=args.sig_intact, total=n_total, misc_thresh=0.00001
    )
    util.count_dict_out(partial_counts_simple, "completeness", total=n_total)

    df_comp = util.count_dict_to_df(
        partial_counts_simple, kind="bead_complete", n_total=n_total
    ).sort_values("name")

    df_pT = util.count_dict_to_df(pT_counts, kind="polyT_after", n_total=n_total)

    df = pd.concat([df_sig, df_concat, df_reprimed, df_comp, df_pT])
    store_results(df, args.stats_out, f"{args.sample_name}.stats.tsv", args.logger)

    # TODO: prefix/suffix counts add up to > 100%. Needs fix
    util.count_dict_out(pT_counts, "polyT after", total=n_total)
    util.count_dict_out(prefixes, "prefixes", total=n_total)
    util.count_dict_out(suffixes, "suffixes", total=n_total)

    # Gather statistics about the parts that make up intact oligos
    qintact, qL, qstarts, qends, qscores = annotation.query_dimensions(
        args.sig_core, substring=True
    )
    # print(qstarts.shape, qintact.shape, qL.shape)
    from collections import Counter

    data = []
    for part, starts, ends, scores in zip(args.sig_core, qstarts.T, qends.T, qscores.T):
        starts_hist = sorted(Counter(starts).items())
        ends_hist = sorted(Counter(ends).items())
        lens_hist = sorted(Counter(ends - starts).items())
        scores_hist = sorted(Counter(scores).items())

        for x, f in starts_hist:
            data.append((args.intact_bead, part, "start", x, f))

        for x, f in ends_hist:
            data.append((args.intact_bead, part, "end", x, f))

        for x, f in lens_hist:
            data.append((args.intact_bead, part, "len", x, f))

        for x, f in scores_hist:
            data.append((args.intact_bead, part, "score", x, f))

    # For reference, also gather statistics for each part of the intact signature,
    # regardless of whether it occurs in the context of an intact signature or not,
    # i.e. what do *all matches* look like
    for part in args.sig_intact:
        qnames, starts, ends, scores, qL = annotation.query_oligo_occurrences(part)
        starts_hist = sorted(Counter(starts).items())
        ends_hist = sorted(Counter(ends).items())
        lens_hist = sorted(Counter(ends - starts).items())
        scores_hist = sorted(Counter(scores).items())

        for x, f in starts_hist:
            data.append(("anywhere", part, "start", x, f))

        for x, f in ends_hist:
            data.append(("anywhere", part, "end", x, f))

        for x, f in lens_hist:
            data.append(("anywhere", part, "len", x, f))

        for x, f in scores_hist:
            data.append(("anywhere", part, "score", x, f))

    df_parts = pd.DataFrame(
        data, columns=["signature", "oligo", "attr", "value", "freq"]
    )
    store_results(
        df_parts, args.stats_out, f"{args.sample_name}.intact_parts.tsv", args.logger
    )

    # output representative examples
    eo_fname = os.path.join(
        util.ensure_path(args.examples_out), f"{args.sample_name}.txt"
    )

    with open(eo_fname, "wt") as eo:
        for signame, sigcount in sorted(sig_counts.items(), key=lambda x: -x[1]):
            try:
                qname, _, _ = next(
                    annotation.filter_signatures(tuple(signame.split(",")))
                )
                eo.write(f"# {signame} n={sigcount}\n{annotation.fmt(qname)}\n")
            except StopIteration:
                args.logger.warning(
                    f"unable to find any reads with signature {signame}"
                )


def get_synth_rates(found_part_counts, all_parts, n_total):
    rates = []

    n0 = n_total
    x = range(1, len(all_parts) + 1)
    for i in x:
        key = tuple(all_parts[:i])
        # print(found_part_counts)
        # print(key, found_part_counts[key], n0, rates)
        rates.append(found_part_counts[key] / float(max(1, n0)))
        # print(i, rates)
        n0 = found_part_counts[key]

    return np.array(rates)


def rep_main(args):
    args = setup_namespace(args)
    args.logger.info(f"generating report plots for '{args.sample_name}'")
    args.logger.debug(f"'bead-related' if we detect: '{args.bead_related}'")

    df = load_results(args.stats_out, f"{args.sample_name}.stats.tsv", args.logger)
    sig_counts = util.count_dict_from_df(df, "signatures")
    util.count_dict_out(
        sig_counts, "signatures", misc_thresh=0.05, total=sig_counts["n_total"]
    )
    n_total = sig_counts["n_total"]
    args.logger.info(f"n_total={n_total}")

    # disentangle bead-related from other signatures
    ov_counts, bead_counts, found_part_counts, core_signature = util.digest_signatures(
        sig_counts,
        args.bead_related,
        args.intact_bead,
        # prefixes=args.prefixes,
        # suffixes=args.suffixes,
    )
    # group low abundance signatures into 'misc' for overview donut plot
    ov_simple, _ = util.count_dict_out(
        ov_counts, "overview counts", misc_thresh=0.01, total=sig_counts["n_total"]
    )
    # group low abundance signatures into 'misc' for bead completeness donut plot
    bead_simple, _ = util.count_dict_out(
        bead_counts, "bead counts", misc_thresh=0.0, total=ov_counts["bead-related"]
    )
    # store the donut plot values in ...report.tsv for cross-sample overviews and such
    df_ov = util.count_dict_to_df(ov_simple, kind="overview")
    df_bead = util.count_dict_to_df(bead_simple, kind="bead_related")
    df_rep = pd.concat([df_ov, df_bead])
    df_rep["sample"] = args.sample_name
    df_rep["signature"] = args.signature
    print(df_rep)
    store_results(df_rep, args.stats_out, f"{args.sample_name}.report.tsv", args.logger)

    # compute and store apparent synthesis rates
    syn_rates = get_synth_rates(found_part_counts, core_signature, n_total)
    df = pd.DataFrame(dict(segment=core_signature, rate=syn_rates))
    df["sample"] = args.sample_name
    store_results(df, args.stats_out, f"{args.sample_name}.synth.tsv", args.logger)

    # prepare for plotting
    del ov_simple["n_total"]
    # assert np.array(list(ov_simple.values())).sum() == n_total
    print(
        f"sum of ov_simple {np.array(list(ov_simple.values())).sum()} n_total={n_total}"
    )

    ov_items = sorted(ov_simple.items(), key=lambda x: -x[1])
    ov_labels = [x[0] for x in ov_items]
    ov_counts = [x[1] for x in ov_items]

    bead_items = sorted(bead_simple.items(), key=lambda x: -x[1])
    bead_labels = [x[0] for x in bead_items]
    bead_counts = [x[1] for x in bead_items]

    # render the donut plots #
    report.plot_results(
        sig_counts,
        ov_labels,
        ov_counts,
        bead_labels,
        bead_counts,
        syn_rates,
        all_parts=core_signature,
        fname=os.path.join(
            util.ensure_path(args.report_out), f"{args.sample_name}.donuts.pdf"
        ),
        suptitle=args.sample_name,
    )

    # render the pre-computed position/score/length histograms as a plot
    report.plot_histograms(
        df=load_results(
            args.stats_out, f"{args.sample_name}.intact_parts.tsv", args.logger
        ),
        fname=os.path.join(
            util.ensure_path(args.report_out), f"{args.sample_name}.hists.pdf"
        ),
        n_total=n_total,
        parts=args.intact_bead.split(","),
    )


def main_edits(args):
    args = setup_namespace(args)

    annotation = ann.AnnotatedSequences(
        args.fname,
        os.path.join(args.annotation_out, f"{args.sample_name}.annotation.tsv"),
        args.sample_name,
        args.blocks,
        min_score=args.min_score,
    )
    n_total = len(annotation.raw_sequences)
    logging.info(
        f"total number of reads in {args.fname} ({args.sample_name}) is {n_total}"
    )

    data = []
    for part in args.sig_intact:
        qmatches = annotation.query_oligo_occurrences(part)
        if len(qmatches[0]) > args.n_samples:
            qmatches = ann.subsample(qmatches, n=args.n_samples)

        nmatch = len(qmatches[0])
        m, ed = ann.align_stats(annotation, args.blocks[part], qmatches)
        for x in np.arange(len(m)):
            # print(part, x, m)
            data.append((part, args.blocks[part], nmatch, x, m[x], ed[x]))

    df = pd.DataFrame(
        data, columns=["oligo", "seq", "nmatch", "pos", "fmatch", "ed_dict"]
    )
    store_results(
        df, args.stats_out, f"{args.sample_name}.oligo_edits.tsv", args.logger
    )

    report.plot_edits(
        df,
        os.path.join(args.report_out, f"{args.sample_name}.oligo_edits.pdf"),
        parts=args.sig_intact,
    )


def main_extract(args):
    sample_name = detect_sample(args)
    blocks = util.load_oligos(args.blocks)

    annotation = ann.AnnotatedSequences(
        args.fname,
        os.path.join(args.annotation_out, f"{sample_name}.annotation.tsv"),
        sample_name,
        blocks,
        min_score=args.min_score,
    )
    n_total = len(annotation.raw_sequences)
    logging.info(f"total number of reads in {args.fname} ({sample_name}) is {n_total}")

    anchor_scores = defaultdict(float)
    barcodes = defaultdict(lambda: "NA")
    umis = defaultdict(lambda: "NA")

    cb_start, cb_end = args.CB.split(",")
    cb_start, cb_end = int(cb_start), int(cb_end)

    umi_start, umi_end = args.UMI.split(",")
    umi_start, umi_end = int(umi_start), int(umi_end)

    if args.barcode_after:
        hits = annotation.query_oligo_occurrences(args.barcode_after)
        for qname, start, end, score, L in zip(*hits):
            a_score = anchor_scores[qname]
            if score > a_score:
                anchor_scores[qname] = score
                seq = annotation.raw_sequences[qname]
                barcodes[qname] = seq[end + cb_start : end + cb_end]
                umis[qname] = seq[end + umi_start : end + umi_end]

    if args.top_barcodes:
        known = set([bc.strip()[::-1] for bc in open(args.top_barcodes).readlines()])
        rev = set([bc[::-1] for bc in known])
        detect = set(barcodes.values())
        nd = len(detect)
        ov = len(detect & known)
        ctrl = len(detect & rev)
        logging.info(
            f"loaded {len(known)} barcodes from '{args.top_barcodes}'. "
            f"{ov} / {nd} detected barcodes overlap ({ov/nd * 100:.3f}%). "
            f"Reverse BC control {ctrl/nd * 100:.3f}%"
        )
    else:
        known = set()

    n = 0
    for qname, sig in annotation.signatures.items():
        sig_str = ",".join(sig)
        if args.sig_include and not (args.sig_include in sig_str):
            continue

        if args.sig_exclude and (args.sig_exclude in sig_str):
            continue

        if args.cDNA_after in sig:
            (
                (cDNA_start, cDNA_end),
                (start_oli, end_oli),
                cDNA,
            ) = annotation.extract_cDNA(
                qname, after_oligo=args.cDNA_after, distal=args.distal
            )
            bc = barcodes[qname]
            if bc not in known:
                bc = bc.lower()
            n += 1
            seq = annotation.raw_sequences[qname]
            header = (
                f">{n}__CB:{bc}__UMI:{umis[qname]}__"
                + f"sig:{','.join(sig)}__cDNA:{cDNA_start}-{cDNA_end}__oli:{start_oli}-{end_oli}__"
                + f"L_read={len(seq)}__L_cDNA={cDNA_end - cDNA_start}"
            )

            # print(header[: (254 - 14)] + " 1:N:0:TCCTGAGC" + f"\n{cDNA}")
            print(header[: (254 - 14)] + " 1:N:0:TCCTGAGC" + f"\n{cDNA}")


def main_overview(args):
    dfs = []
    for fname in list(args.reports):
        print(f"loading {fname}")
        df = pd.read_csv(fname, sep="\t", index_col=0)
        dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)

    dfsig = df[["sample", "signature"]].drop_duplicates().set_index("sample")
    print(dfsig)
    df = df[["name", "count", "sample"]].pivot(
        index="sample", columns="name", values="count"
    )  # .query("kind == 'overview'")
    df = df.fillna(0).div(df["n_total"], axis=0)
    df *= 100  # we'd like percentages

    repriming = [
        "TSO,TSO_RC",
        "dN-SMRT,dN-SMRT_RC",
    ]
    concatenation = [c for c in df.columns if "+" in c]
    bead = ["complete", "only_bead_start", "missing_polyT", "missing_OP1"][::-1]

    # # avoid crash if columns are missing
    for r in repriming + concatenation + bead:
        if r not in df.columns:
            df[r] = 0

    # print(df)
    # print(f"concat columns {concatenation}")
    # print(f"bead columns {bead}")
    df["reprimed"] = df[repriming].sum(axis=1)
    # df["complete"] = np.nan_to_num(df["complete"], nan=0.0)
    df["concat"] = df[concatenation].sum(axis=1)
    df["bead_related"] = np.nan_to_num(df[bead].sum(axis=1), nan=0.0)

    def ds(row):
        sig = dfsig.loc[row.name]
        res = (sig == "dropseq") * row.complete
        return res

    df["dropseq"] = df[["complete"]].apply(ds, axis=1)
    # df["complete"] = df[""]
    df["incomplete"] = df["bead-related"] - df["complete"] - df["dropseq"]
    df["non_bead"] = 100 - df["bead-related"]
    df["fidelity"] = 100 * df["complete"] / df["bead-related"]
    df = df.fillna(0)
    # print(df)
    if args.csv_out:
        df.to_csv(args.csv_out, float_format="%.2f", sep="\t")

    def clean(txt):
        txt = os.path.basename(txt)
        t = (
            txt.replace("source/", "")
            .replace("sts_", "")
            .replace("pb_", "")
            .replace("ds_", "")
            .replace(".fq", "")
            .replace(".bam", "")
            .replace("lima.", "")
        )

        if t.count("_") > 1:
            t = "_".join(t.split("_")[:2])

        return t

    df = df.reset_index()
    # df = df.sort_values('bead_related')
    df = df.sort_values("sample")

    def guess_rRNA_file(path):
        # print("guessrRNA raw path", path)
        name = os.path.basename(path).replace(".summary", ".rRNA")

        if args.rRNA_same_place:
            place = os.path.dirname(path)
        else:
            place = args.rRNA

        return [
            os.path.join(place, name.replace(".fq", ".txt")),
            os.path.join(place, name.replace(".fq", ".txt")).replace(
                ".rRNA.tsv", ".txt"
            ),
            os.path.join(place, name.replace(".fq", ".txt")).replace(
                ".rRNA.tsv", ".rRNA.txt"
            ),
            os.path.join(place, name.replace(".bam", ".txt").replace("lima.", "")),
            os.path.join(
                place, name.replace(".bam", ".txt").replace("lima.", "")
            ).replace(".rRNA.tsv", ".txt"),
            os.path.join(
                place, name.replace(".bam", ".txt").replace("lima.", "")
            ).replace(".rRNA.tsv", ".rRNA.txt"),
        ]

    # rRNA_fracs = []
    # for row in df[["stats_file", "N_reads"]].itertuples():
    #     rcount = np.nan
    #     for fname in guess_rRNA_file(row.stats_file):
    #         print(fname)
    #         try:
    #             rcount = int(open(fname).read())
    #         except (FileNotFoundError, ValueError):
    #             pass
    #         else:
    #             break
    #     if rcount == np.nan:
    #         raise ValueError

    #     rRNA_fracs.append(100.0 * rcount / row.N_reads)

    # df["rRNA"] = rRNA_fracs
    # print(df[['qfa', 'rRNA']])

    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    plt.rcParams.update({"font.size": 8})

    def make_bars(
        ax, df, kinds, labels, cmap=plt.get_cmap("tab10"), w=0.9, colors=None
    ):
        n = len(kinds)
        if colors is None:
            colors = cmap(np.linspace(0, 1, n))

        x = np.arange(len(df)) - w / 2.0
        y0 = np.zeros(len(x), dtype=float)
        for kind, label, color in zip(kinds, labels, colors):
            y = np.nan_to_num(df[kind], nan=0.0)
            # print(kind)
            # print(y)
            ax.bar(x, y, bottom=y0, label=label, width=w, color=color)
            y0 += y

        ax.set_ylabel("fraction of library")
        ax.set_xticks(x)
        labels = df["sample"]  # [clean(fq) for fq in df['qfa']]
        ax.set_xticklabels(labels, rotation=90)
        ax.set_ylim(0, 100)

    marie = [
        "non_bead",
        "incomplete",
        "dropseq",
        "complete",
    ]
    marie_colors = ["gray", "royalblue", "green", "gold"]

    w = max(8 / 25.0 * len(df), 5)
    if args.multi_page:
        pdf = PdfPages(args.breakdown)
        fig, ax1 = plt.subplots(1, figsize=(w, 8))
    else:
        fig, (ax1, ax2) = plt.subplots(2, figsize=(w, 8), sharex=True)

    make_bars(
        ax1,
        df,
        marie,
        labels=marie,
        colors=marie_colors,
    )
    ax1.legend(title="Marie-stats", ncol=len(marie))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax2 = plt.subplots(1, figsize=(w, 4))

    make_bars(ax2, df, ["fidelity"], labels=["bead fidelity"])
    ax2.set_ylabel("bead fidelity")
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        pdf.close()
    else:
        fig.tight_layout()
        plt.savefig(args.breakdown)

    plt.close()

    if args.multi_page:
        pdf = PdfPages(args.output)
        fig, ax1 = plt.subplots(1, figsize=(w, 4))
    else:
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, figsize=(w, 12), sharex=True)

    # print("bead related", bead)
    make_bars(ax1, df, bead, labels=bead)
    ax1.legend(title="bead-related", ncol=len(bead))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax2 = plt.subplots(1, figsize=(w, 4))

    # print("repriming events", repriming)
    make_bars(
        ax2,
        df,
        repriming,
        labels=[r.split(",")[0] for r in repriming],
        cmap=plt.get_cmap("tab20c"),
    )
    ax2.legend(title="repriming", ncol=len(repriming))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax3 = plt.subplots(1, figsize=(w, 4))

    # print("concat events", concatenation)
    make_bars(ax3, df, concatenation, labels=concatenation, cmap=plt.get_cmap("tab20b"))
    ax3.legend(title="concatamers", ncol=len(concatenation))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax4 = plt.subplots(1, figsize=(w, 4))

    # make_bars(
    #     ax4,
    #     df,
    #     [
    #         "rRNA",
    #     ],
    #     labels=["rRNA"],
    #     cmap=plt.get_cmap("tab20c"),
    # )
    ax4.legend(title="human rRNA", ncol=1)
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        pdf.close()
    else:
        fig.tight_layout()
        plt.savefig(args.output)

    plt.close()
    return df


def prepare_parser():
    parser = argparse.ArgumentParser(prog="longread")
    parser.add_argument(
        "--sample",
        help="sample name (default=autodetect from fname)",
        default=None,
    )
    parser.add_argument(
        "--config",
        help="YAML describing expected longread signature (default=./longread.yaml if detected or built-in)",
        default="longread.yaml",
    )
    parser.add_argument(
        "--signature",
        help="expected long read signature (e.g. dropseq, noUMI, withUMI,...)",
        default=None,
    )
    parser.add_argument(
        "--blocks",
        default="",
        help="FASTA file with known oligo sequences (default=built-in)",
    )
    parser.add_argument(
        "--cache",
        default="./cache/",
        help="path to alignment caches (default=./cache/)",
    )
    parser.add_argument(
        "--annotation-out",
        default="./annotation/",
        help="path to store annotation data in (default=./annotation/)",
    )
    parser.add_argument(
        "--examples-out",
        help="path to store annotated example read sequences",
        default="./examples/",
    )
    parser.add_argument(
        "--stats-out",
        help="path to store statistics (pandas dataframes)",
        default="./stats/",
    )
    parser.add_argument(
        "--report-out", help="path to render graphical reports in", default="./reports/"
    )
    parser.add_argument(
        "--parallel",
        default=16,
        type=int,
        help="number of parallel processes (default=16)",
    )
    parser.add_argument(
        "--min-score",
        default=0.6,
        type=float,
        help="minimal match alignment score to consider a match for annotation, relative to its size (default=0.6)",
    )
    parser.add_argument(
        "--debug", default=False, action="store_true", help="activate debug output"
    )

    ## sub-parser setup ##
    subparsers = parser.add_subparsers(help="sub-command help", dest="subcmd")
    aln_parser = subparsers.add_parser(
        "align", help="align PacBio reads against oligos"
    )
    aln_parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    ann_parser = subparsers.add_parser(
        "annotate", help="create annotation from detected oligo matches"
    )
    ann_parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    ann_parser.add_argument(
        "--polyT",
        help="name of oligo-block that matches polyT (default='polyT')",
        default="polyT",
    )

    rep_parser = subparsers.add_parser("report", help="create PDF/PNG reports")

    ed_parser = subparsers.add_parser(
        "edits", help="gather mismatch and indel stats for oligo matches"
    )
    ed_parser.add_argument(
        "--n-samples",
        type=int,
        default=1000,
        help="number of sample alignments to gather edit statistics from (default=1000)",
    )
    ed_parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )

    xt_parser = subparsers.add_parser(
        "extract", help="extract sequences from the long read"
    )
    xt_parser.add_argument(
        "--barcode-after",
        type=str,
        default="bead_start",
        help="name of anchor match after which the barcodes follow (default='bead_start')",
    )
    xt_parser.add_argument(
        "--CB",
        type=str,
        default="8,20",
        help="bases downstream of anchor match at which the cell barcode starts and ends. Default='8,20'",
    )
    xt_parser.add_argument(
        "--UMI",
        type=str,
        default="0,8",
        help="bases downstream of anchor match at which the UMI starts and ends. Default='0,8'",
    )
    xt_parser.add_argument(
        "--cDNA-after",
        type=str,
        default="bead_start",
        help="excise sequence between this oligo and the last oligo match (if it exists)",
    )
    xt_parser.add_argument(
        "--sig-include",
        type=str,
        default="",
        help="extract cDNA only from long reads INCLUDING this substring in the signature",
    )
    xt_parser.add_argument(
        "--sig-exclude",
        type=str,
        default="",
        help="extract cDNA only from long reads EXCLUDING this substring in the signature",
    )
    xt_parser.add_argument(
        "--distal",
        default=150,
        type=int,
        help="number of nt from end of sequence that are considered for oligo matches when extracting cDNA",
    )
    xt_parser.add_argument(
        "--top-barcodes",
        type=str,
        default="",
        help="path to text file with known barcodes (e.g. from Illumina)",
    )

    ov_parser = subparsers.add_parser(
        "overview", help="make overview plots across samples"
    )
    ov_parser.add_argument("reports", nargs="+", help="sample reports to aggregate")
    ov_parser.add_argument(
        "--output", default="pb_overview.pdf", help="path/name of detailed report PDF"
    )
    ov_parser.add_argument(
        "--csv-out", default="all_pb_stats.csv", help="path/name of detailed report PDF"
    )
    ov_parser.add_argument(
        "--breakdown",
        default="bead_overview.pdf",
        help="path/name of bead report (Marie style) PDF",
    )
    # ov_parser.add_argument(
    #     "--rRNA",
    #     default="rRNA/",
    #     help="path to search for rRNA counts corresponding to samples",
    # )
    # ov_parser.add_argument(
    #     "--rRNA-same-place",
    #     default=False,
    #     action="store_true",
    #     help="If set, look for rRNA txt file with same sample name in same directory",
    # )
    ov_parser.add_argument(
        "--multi-page",
        default=False,
        action="store_true",
        help="If set, generate multiple PDF pages instead of subplots",
    )

    parser.set_defaults(func=lambda args: parser.print_help())
    aln_parser.set_defaults(func=aln_main)
    ann_parser.set_defaults(func=ann_main)
    rep_parser.set_defaults(func=rep_main)
    ed_parser.set_defaults(func=main_edits)
    xt_parser.set_defaults(func=main_extract)
    ov_parser.set_defaults(func=main_overview)

    return parser


def cmdline():
    import logging

    logging.basicConfig(level=logging.INFO)

    parser = prepare_parser()
    args = parser.parse_args()
    return args.func(args)



from .cmdline import cmdline
from .dge import calculate_adata_metrics,\
    calculate_shannon_entropy_scompression, dge_to_sparse_adata,\
    attach_barcode_file, parse_barcode_file, load_external_dge




import pysam
import datetime
import argparse
import numpy as np

counted_regions = ['UTR', 'CODING']

def select_alignment(alignments):
    read_names = [aln.query_name for aln in alignments]
    if read_names.count(read_names[0]) != len(read_names):
        print(read_names)
        raise Exception(f'input alignments do not come from the same read')

    def is_exonic(aln):
        if not aln.has_tag('XF'):
            return False

        return aln.get_tag('XF') in counted_regions

    alignments_are_exonic = np.array([is_exonic(aln) for aln in alignments])

    exonic_ix = np.where(alignments_are_exonic  == True)[0]

    num_exonic = exonic_ix.shape[0]

    if num_exonic == 1:
        # if only one exonic reads from the group
        # return the exonic indices
        return alignments[exonic_ix[0]]
    else:
        return None

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Filter out ambiguous multi-mapper reads')

    parser.add_argument('--in-bam', help='input bam')
    parser.add_argument('--out-bam', help='output bam')

    args = parser.parse_args()
    print(args)

    bam_in = pysam.AlignmentFile(args.in_bam, "rb")

    bam_out = pysam.AlignmentFile(args.out_bam, 'wb', header= bam_in.header)
    counter = 0
    start_time = datetime.datetime.now()

    multi_mappers = []
    
    for aln in bam_in.fetch(until_eof=True):
        counter = counter + 1

        if counter % 1000000 == 0:
            finish_time = datetime.datetime.now()
            delta_seconds = (finish_time - start_time).seconds
            # restart time
            start_time = finish_time
            print(f'Processed 1 millon records in {delta_seconds} seconds, total records processed {counter}. current time: {finish_time}')

        mapped_number = aln.get_tag('NH')

        if mapped_number == 1:
            bam_out.write(aln)
        else:
            if len(multi_mappers) < (mapped_number - 1):
                # still some multimappers missing. we need to add the alignments 
                # until the last one to the list
                multi_mappers.append(aln)
            else:
                # add the last alignment
                multi_mappers.append(aln)
                # decide which, if any, to keep
                aln_to_keep = select_alignment(multi_mappers)

                if aln_to_keep is not None:
                    # set aln secondary flag to 0, so that it is flagged as primary
                    # secondary flag is at 0x100, so 8th bit (starting from 0)
                    aln_to_keep.flag = aln_to_keep.flag & ~(1<<8)
                    bam_out.write(aln_to_keep)
                    
                # reset multimapper list
                multi_mappers = []



__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

import shelve
import os
import logging
import numpy as np
import pandas as pd
from Bio import pairwise2
from spacemake.util import ensure_path, read_fq

"""
A module that uses Smith & Waterman alignment from Biopython to detect
matches of pre-defined sequence blocks (oligos) in PacBio long reads.
Matches are stored using the shelve module for fast access in in-depth
analyses (CachedAlignments class).
Overlapping alignments and co-occurence pattern annotation are handled by
the MultiAlignments class.
"""


def align(ref, query, aln_scores=(2, -1.5, -3, -1, -3, -1), min_score=0.4):
    L = min(len(ref), len(query))
    ms = max(int(min_score * L * aln_scores[0]), 14)  # at least 7 matches
    return [
        r for r in pairwise2.align.localmd(ref, query, *aln_scores) if r.score >= ms
    ]


def print_aln(aln):
    s = pairwise2.format_alignment(*aln, full_sequences=True)
    print(s)
    return s


def non_overlapping_hits(seq, oligo, min_size=7):
    oli_hits = np.zeros(len(seq), dtype=int)
    hit_scores = np.zeros(len(seq))
    #
    for res in align(seq, oligo):
        # print(res.score, res.start, res.end)
        start = res.start
        end = res.end + 1
        scores = hit_scores[start:end]
        hits = oli_hits[start:end]
        oli_hits[start:end] = np.where(scores < res.score, 1, hits)
        hit_scores[start:end] = np.where(scores < res.score, res.score, scores)
    #
    positions = (oli_hits > 0).nonzero()[0]
    if len(positions):
        start = positions[0]
        last = start
        for i in positions[1:]:
            if i > last + 1:
                if last - start > min_size:
                    yield start, last, hit_scores[start:last].max()
                start = i
            last = i
        #
        if last - start > min_size:
            yield start, last, hit_scores[start:last].max()


class CachedAlignments:
    def __init__(self, sample_name, oligo_name, oligo_seq, path="./cache/"):
        """
        Sets up transparent caching of alignments. A shelve DB will be
        created in {path}/{sample_name}__{oligo_name}.<db suffixes>, which allows to
        concurrently align different oligos against the same reads.
        """
        self.logger = logging.getLogger("CachedAlignments")
        self.logger.info(f"initializing for sample={sample_name} oligo={oligo_name}")
        self.sample_name = sample_name
        self.oligo_name = oligo_name
        self.oligo_seq = oligo_seq
        self.oli_lengths = {"NA": 0}
        for oli, seq in zip(self.oligo_name, self.oligo_seq):
            self.oli_lengths[oli] = float(len(seq))

        self.sname = os.path.join(ensure_path(path), f"{sample_name}__{oligo_name}")
        self.shelf = shelve.open(self.sname)
        self.logger.info(f"found {len(self.shelf)} cached entries")

    def align_or_load(self, seq_name, seq):
        if not seq_name in self.shelf:
            hits = list(non_overlapping_hits(seq, self.oligo_seq))
            self.shelf[seq_name] = hits

        return self.shelf[seq_name]

    def close(self):
        self.shelf.sync()
        self.shelf.close()


class MultiAlignments:
    def __init__(self, sample_name, oligo_dict, path="./cache/"):
        """
        Combine alignments of multiple oligos against a set of reads and
        annotate occurence patterns. Caches will be built/opened from
        {path}/{sample_name}__{oligo_name} for all oligos in oligo_dict,
        which maps oligo_name -> oligo_sequence.
        """
        self.logger = logging.getLogger("MultiAlignments")
        self.sample_name = sample_name
        self.oligo_dict = oligo_dict
        self.oligo_names = sorted(oligo_dict.keys())
        self.aln_caches = {}
        for o in self.oligo_names:
            aln = CachedAlignments(sample_name, o, self.oligo_dict[o])
            self.aln_caches[o] = aln

    def annotate(self, seq_name, seq, min_size=7):
        """
        Iterates over all oligo matches for a desired read sequence and yields
        (start, end, label, score) annotations
        """
        labels = [
            "NA",
        ] + self.oligo_names
        L = len(seq)
        oli_hits = np.zeros(L, dtype=int)
        hit_scores = np.zeros(L)

        for i, o in enumerate(self.oligo_names):
            aln = self.aln_caches[o]
            for start, end, score in aln.align_or_load(seq_name, seq):
                scores = hit_scores[start:end]
                hits = oli_hits[start:end]
                oli_hits[start:end] = np.where(scores < score, i + 1, hits)
                hit_scores[start:end] = np.where(scores < score, score, scores)

        positions = (oli_hits > 0).nonzero()[0]
        if len(positions):
            start = positions[0]
            prev = oli_hits[start]
            last = start
            for i in positions[1:]:
                curr = oli_hits[i]
                if curr != prev or i > last + 1:
                    if last - start > min_size:
                        label = labels[prev]
                        score = hit_scores[
                            start:last
                        ].max()  # * (last - start) / self.oli_lengths[label]
                        yield start, last, label, score
                    start = i
                last = i
                prev = curr
            #
            if last - start > min_size:
                label = labels[prev]
                score = hit_scores[
                    start:last
                ].max()  # * (last - start) / self.oli_lengths[label]
                yield start, last, labels[prev], score


def align_one_oligo_one_read(oligo_seq, qname, qseq):
    hits = list(non_overlapping_hits(qseq, oligo_seq))
    return qname, hits


def align_one_oligo(oligo_name, oligo_seq, sample_name, fastq_path, n_proc=None):
    import multiprocessing as mp

    aln = CachedAlignments(sample_name, oligo_name, oligo_seq)
    n = 0
    job_list = []
    for name, seq, qual in read_fq(fastq_path):
        if not name in aln.shelf:
            job_list.append((oligo_seq, name, seq))

        # aln.align_or_load(name, seq)
        n += 1

    if job_list:
        logging.info(
            f"about to align {len(job_list)} reads with oligo {oligo_name} using {n_proc} processes"
        )
        pool = mp.Pool(n_proc)
        for qname, hits in pool.starmap(
            align_one_oligo_one_read, job_list, chunksize=10
        ):
            aln.shelf[qname] = hits
    else:
        logging.info(
            f"every read was already aligned against {oligo_name}. Nothing to do."
        )

    return sample_name, oligo_name, n


def fill_caches(
    fastq_path, sample_name, oligo_dict, path="./cache/", n_proc=None
):
    """
    Utility function to fill the shelves by running alignments of different oligos.
    Not parallelized
    """
    # import multiprocessing as mp
    for oligo_name, oligo_seq in sorted(oligo_dict.items()):
        sample_name, oligo_name, n_reads = align_one_oligo(
            oligo_name, oligo_seq, sample_name, fastq_path, n_proc=n_proc
        )
        print(f"{sample_name}: aligned {n_reads} against {oligo_name}")


def annotate(fastq_path, sample_name, oligo_dict, path="./cache/"):
    multi = MultiAlignments(sample_name, oligo_dict, path=path)
    data = []
    for name, seq, qual in read_fq(fastq_path):
        for start, end, label, score in multi.annotate(name, seq):
            data.append((sample_name, name, len(seq), label, start, end, score))

    df = pd.DataFrame(
        data, columns=["sample_name", "qname", "L", "oligo", "start", "end", "score"]
    )
    return df



import pandas as pd
import os
import yaml
import math
import argparse
import datetime
import re
import logging

from spacemake.errors import *
from spacemake.config import ConfigFile
from spacemake.util import message_aggregation, assert_file, str_to_list
from typing import List, Dict

logger_name = "spacemake.project_df"


def get_project_sample_parser(allow_multiple=False, prepend="", help_extra=""):
    """
    Return a parser for project_id's and sample_id's

    :param allow_multiple: if true, we allow multiple projects and samples,
        and the parser will have a `--project_id_list` and a `--sample_id_list`
        parameter
    :param prepend: a string that will be prepended before the arguments
    :param help_extra: extra help message
    ...
    :return: a parser object
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    project_argument = "project_id"
    sample_argument = "sample_id"
    required = True
    nargs = None
    default = None

    if allow_multiple:
        project_argument = f"{project_argument}_list"
        sample_argument = f"{sample_argument}_list"
        required = False
        nargs = "*"
        default = []

    parser.add_argument(
        f"--{prepend}{project_argument}",
        type=str,
        required=required,
        nargs=nargs,
        default=default,
        help=f"{project_argument} {help_extra}",
    )

    parser.add_argument(
        f"--{prepend}{sample_argument}",
        type=str,
        required=required,
        nargs=nargs,
        default=default,
        help=f"{sample_argument} {help_extra}",
    )

    return parser


def get_add_sample_sheet_parser():
    """
    Returns parser for sample sheet addition

    :return: parser
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="add a new sample sheet to the samples",
        add_help=False,
    )

    parser.add_argument(
        "--sample_sheet",
        type=str,
        help="the path to the Illumina sample sheet",
        required=True,
    )
    parser.add_argument(
        "--basecalls_dir",
        type=str,
        help="path to the basecalls directory",
        required=True,
    )

    return parser

def get_sample_main_variables_parser(
        species_required=False,
        main_variables=['barcode_flavor', 'species', 'puck', 'run_mode']
    ):
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    if 'barcode_flavor' in main_variables:
        parser.add_argument(
            "--barcode_flavor", type=str, help="barcode flavor for this sample"
        )

    if 'species' in main_variables:
        parser.add_argument(
            "--species",
            type=str,
            help="add the name of the species for this sample",
            required=species_required,
        )

    if 'puck' in main_variables:
        parser.add_argument(
            "--puck",
            type=str,
            help="name of the puck for this sample. if puck contains a \n"
            + "`barcodes` path to a coordinate file, those coordinates\n"
            + " will be used when processing this sample. if \n"
            + " not provided, a default puck will be used with \n"
            + "width_um=3000, spot_diameter_um=10",
        )

        parser.add_argument(
            "--puck_id", type=str, help="puck_id of the sample to be added/update"
        )

    parser.add_argument(
        "--puck_barcode_file",
        type=str,
        help="the path to the file contining (x,y) positions of the barcodes",
    )


    if 'run_mode' in main_variables:
        parser.add_argument(
            "--run_mode",
            type=str,
            nargs="+",
            help="run_mode names for this sample.\n"
            + "the sample will be processed using the provided run_modes.\n"
            + "for merged samples, if left empty, the run_modes of the \n"
            + "merged (input) samples will be intersected.\n",
        )

    return parser

def get_sample_extra_info_parser():
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    parser.add_argument(
        "--investigator",
        type=str,
        help="add the name of the investigator(s) responsible for this sample",
    )

    parser.add_argument("--experiment", type=str, help="description of the experiment")

    parser.add_argument(
        "--sequencing_date",
        type=datetime.date.fromisoformat,
        help="sequencing date of the sample. format: YYYY-MM-DD",
    )

    return parser

def get_data_parser(reads_required=False):
    """
    Returns a parser which contain extra arguments for a given sample.
    The returned parser will contain the --R1, --R2, --longreads,
    --longread-signature, --barcode_flavor, --species, --puck,  --puck_id,
    --puck_barcode_file, --investigator, --experiment, --sequencing_date,
    --run_mode arguments.

    :param species_required: if true, the --species argument will be required
        during parsing.
    :param reads_required: if true, --R1, --R2, and --longreads arguments will be
        required during parsing.
    """
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    parser.add_argument(
        "--R1",
        type=str,
        help=".fastq.gz file path to R1 reads",
        nargs="+",
        required=reads_required,
    )

    parser.add_argument(
        "--R2",
        type=str,
        help=".fastq.gz file path to R2 reads",
        nargs="+",
        required=reads_required,
    )

    parser.add_argument(
        "--dge",
        type=str,
        help="Path to dge matrix. spacemake can also handle already processed"
        + " digital expression data",
        required=reads_required,
    )

    parser.add_argument(
        "--longreads",
        type=str,
        help="fastq(.gz)|fq(.gz)|bam file path to pacbio long reads for library debugging",
        required=reads_required,
    )

    parser.add_argument(
        "--longread-signature",
        type=str,
        help="identify the expected longread signature (see longread.yaml)",
    )

    return parser


def get_set_remove_variable_subparsers(
    parent_parser, variable_name, func, prepend="", allow_multiple=False
):
    """get_set_remove_variable_subparsers.

    :param parent_parser:
    :param variable_name:
    :param func:
    :param prepend:
    :param allow_multiple:
    """
    if allow_multiple:
        nargs = "+"
    else:
        nargs = None

    def get_action_parser(action):
        """get_action_parser.

        :param action:
        """
        action_parser = parent_parser.add_parser(
            f"{action}_{variable_name}",
            parents=[get_project_sample_parser(allow_multiple=True)],
            description=f"{action} {variable_name} for several projects/samples",
            help=f"{action} {variable_name} for several projects/samples",
        )
        action_parser.add_argument(
            f"--{variable_name}", type=str, required=True, nargs=nargs
        )
        action_parser.set_defaults(func=func, variable=variable_name, action=action)

        return action_parser

    set_parser = get_action_parser("set")

    if allow_multiple:
        set_parser.add_argument("--keep_old", action="store_true")

        remove_parser = get_action_parser("remove")


def get_action_sample_parser(parent_parser, action, func):
    """get_action_sample_parser.

    :param parent_parser:
    :param action:
    :param func:
    """
    if action not in ["add", "update", "delete", "merge"]:
        raise ValueError(f"Invalid action: {action}")

    if action == "merge":
        parser_name = "merge_samples"
        msg = "merge samples"
        parents = [
            get_project_sample_parser(prepend="merged_",
                help_extra='of the newly created merged sample'),
            get_project_sample_parser(allow_multiple=True,
                help_extra='of the samples to be merged'),
        ]
    else:
        parser_name = f"{action}_sample"
        msg = f"{action} a sample"
        parents = [get_project_sample_parser()]

    if action == "add":
        # add arguments for species, run_mode, barcode_flavor and puck
        parents.append(
            get_sample_main_variables_parser(
                species_required=True,
            )
        )
        # add arguments for R1/R1, dge, longread
        parents.append(
            get_data_parser()
        )
        # add arguments for extra sample info
        parents.append(
            get_sample_extra_info_parser()
        )
    elif action == 'update':
        # add main variables parser
        parents.append(
            get_sample_main_variables_parser()
        )

        # add possibility to add extra info
        parents.append(
            get_sample_extra_info_parser()
        )
    elif action == "merge":
        # add main variables parser
        # when merging, only let the user overwrite puck and run_mode
        parents.append(
            get_sample_main_variables_parser(
                main_variables=['run_mode', 'puck'],
            )
        )

        # add possibility to add extra info
        parents.append(
            get_sample_extra_info_parser()
        )

    sample_parser = parent_parser.add_parser(
        parser_name, description=msg, help=msg, parents=parents
    )
    sample_parser.set_defaults(func=func, action=action)


def setup_project_parser(pdf, attach_to):
    """setup_project_parser.

    :param pdf:
    :param attach_to:
    """
    parser = attach_to.add_parser(
        "projects",
        help="manage projects and samples",
        description="Using one of the subcommands specified below, it is possible to"
        + " add/update/remove projects and their settings",
    )
    subparsers = parser.add_subparsers()

    help_desc = {
        "add_sample_sheet": "add projects and samples from Illumina sample sheet",
        "add_samples_from_yaml": "add several samples at once from a .yaml file",
        "merge_samples": "merge several samples into one. "
        + "samples need to have the same species",
        "list": "list several project(s) and sample(s) and their settings",
    }

    # ADD/UPDATE/DELETE/MERGE sample(s)
    # get parsers for adding, deleting and updating a sample
    for action in ["add", "update", "delete"]:
        get_action_sample_parser(
            subparsers,
            action,
            # attach the function to the parser
            func=lambda args: add_update_delete_sample_cmdline(pdf, args),
        )

    # get parser for merging
    get_action_sample_parser(
        subparsers, "merge", func=lambda args: merge_samples_cmdline(pdf, args)
    )

    # LIST PROJECTS
    # always show these variables
    always_show = [
        "puck_id",
        "species",
        "investigator",
        "sequencing_date",
        "experiment",
        "run_mode",
        "barcode_flavor",
    ]
    remaining_options = [
        x for x in pdf.project_df_default_values.keys() if x not in always_show
    ]
    list_projects = subparsers.add_parser(
        "list",
        description=help_desc["list"],
        help=help_desc["list"],
        parents=[
            get_project_sample_parser(
                allow_multiple=True,
                help_extra="subset the data. If none provided, all will be displayed",
            )
        ],
    )
    list_projects.add_argument(
        "--variables",
        help="which extra variables to display per sample? "
        + f"{always_show} will always be shown.",
        choices=remaining_options,
        default=[],
        nargs="*",
    )
    list_projects.set_defaults(
        func=lambda args: list_projects_cmdline(pdf, args), always_show=always_show
    )

    # ADD SAMPLE SHEET
    sample_add_sample_sheet = subparsers.add_parser(
        "add_sample_sheet",
        description=help_desc["add_sample_sheet"],
        help=help_desc["add_sample_sheet"],
        parents=[get_add_sample_sheet_parser()],
    )
    sample_add_sample_sheet.set_defaults(
        func=lambda args: add_sample_sheet_cmdline(pdf, args)
    )

    # ADD SAMPLES FROM YAML
    sample_add_samples_yaml = subparsers.add_parser(
        "add_samples_from_yaml",
        description=help_desc["add_samples_from_yaml"],
        help=help_desc["add_samples_from_yaml"],
    )
    sample_add_samples_yaml.add_argument(
        "--samples_yaml",
        type=str,
        required=True,
        help="path to the .yaml file containing sample info",
    )
    sample_add_samples_yaml.set_defaults(
        func=lambda args: add_samples_from_yaml_cmdline(pdf, args)
    )

    # get set/remove parser for each main variable
    # this will add parser for:
    # setting/removing run_mode
    # setting species
    # setting puck
    # setting barcode_flavor
    for main_var_sg, main_var_type in ConfigFile.main_variable_sg2type.items():
        allow_multiple = False
        if isinstance(main_var_type, str) and main_var_type.endswith("_list"):
            allow_multiple = True

        get_set_remove_variable_subparsers(
            subparsers,
            variable_name=main_var_sg,
            func=lambda args: set_remove_variable_cmdline(pdf, args),
            allow_multiple=allow_multiple,
        )

    return parser


@message_aggregation(logger_name)
def add_update_delete_sample_cmdline(pdf, args):
    """add_update_delete_sample_cmdline.

    :param pdf:
    :param args:
    """
    action = args["action"]

    # remove the action from args
    del args["action"]

    if action == "add" or action == "update":
        func = lambda **kwargs: pdf.add_update_sample(action=action, **kwargs)
    elif action == "delete":
        func = pdf.delete_sample

    sample = func(**args)
    pdf.dump()


@message_aggregation(logger_name)
def add_sample_sheet_cmdline(pdf, args):
    """add_sample_sheet_cmdline.

    :param pdf:
    :param args:
    """
    pdf.add_sample_sheet(args["sample_sheet"], args["basecalls_dir"])

    pdf.dump()


@message_aggregation(logger_name)
def add_samples_from_yaml_cmdline(pdf, args):
    """add_samples_from_yaml_cmdline.

    :param pdf:
    :param args:
    """
    pdf.add_samples_from_yaml(args["samples_yaml"])

    pdf.dump()


@message_aggregation(logger_name)
def set_remove_variable_cmdline(pdf, args):
    """set_remove_variable_cmdline.

    :param pdf:
    :param args:
    """
    variable_name = args["variable"]
    action = args["action"]

    variable_key = args[variable_name]

    pdf.set_remove_variable(
        variable_name=variable_name,
        variable_key=variable_key,
        action=action,
        project_id_list=args["project_id_list"],
        sample_id_list=args["sample_id_list"],
        keep_old=args.get("keep_old", False),
    )

    pdf.dump()


@message_aggregation(logger_name)
def merge_samples_cmdline(pdf, args):
    """merge_samples_cmdline.

    :param pdf:
    :param args:
    """
    pdf.merge_samples(**args)

    pdf.dump()


@message_aggregation(logger_name)
def list_projects_cmdline(pdf, args):
    """list_projects_cmdline.

    :param pdf:
    :param args:
    """
    projects = args["project_id_list"]
    samples = args["sample_id_list"]
    variables = args["always_show"] + args["variables"]

    df = pdf.df
    logger = logging.getLogger(logger_name)

    if projects != [] or samples != []:
        df = df.query("project_id in @projects or sample_id in @samples")
        logger.inof(f"listing projects: {projects} and samples: {samples}")
    else:
        logger.info("listing all projects and samples")

    logger.info(f"variables used: {variables}")

    # print the table
    logger.info(df.loc[:, variables].__str__())


class ProjectDF:
    """
    ProjectDF: class responsible for managing spacemake projects.

    :param file_path: path to the project_df.csv file, where we will save the project list.
    :param config: config file object
    :type config: ConfigFile
    :param df: A pandas dataframe, containing one row per sample
    :type df: pd.DataFrame
    """

    # default values of the project dataframe columns
    project_df_default_values = {
        "puck_id": "no_optical_puck",
        "sample_sheet": None,
        "species": None,
        "demux_barcode_mismatch": 1,
        "demux_dir": None,
        "basecalls_dir": None,
        "R1": None,
        "R2": None,
        "longreads": None,
        "longread_signature": None,
        "investigator": "unknown",
        "sequencing_date": "unknown",
        "experiment": "unknown",
        "puck_barcode_file": None,
        "run_mode": ["default"],
        "barcode_flavor": "default",
        "is_merged": False,
        "merged_from": [],
        "puck": "default",
        "dge": None,
    }

    def __init__(self, file_path, config: ConfigFile = None):
        """__init__.

        :param file_path: path to pandas data frame (saved as .csv)
        :param config: ConfigFile
        :type config: ConfigFile
        """
        self.file_path = file_path
        self.config = config

        if os.path.isfile(file_path):
            df = pd.read_csv(
                file_path,
                index_col=["project_id", "sample_id"],
                converters={"run_mode": eval, "merged_from": eval},
                na_values=["None", "none"],
            )

            # replacing NaN with None
            df = df.where(pd.notnull(df), None)

            # convert R1/R2 to list, if they are stored as string
            df.R1 = df.R1.apply(str_to_list)
            df.R2 = df.R2.apply(str_to_list)

            project_list = []
            # required if upgrading from pre-longread tree
            if not "longreads" in df.columns:
                df["longreads"] = None

            if not "longread_signature" in df.columns:
                df["longread_signature"] = None

            # update with new columns, if they exist.
            for ix, row in df.iterrows():
                s = pd.Series(self.project_df_default_values)
                s.update(row)
                s.name = row.name
                project_list.append(s)

            self.df = pd.concat(project_list, axis=1).T
            self.df.is_merged = self.df.is_merged.astype(bool)
            self.df.index.names = ["project_id", "sample_id"]

            # dump the result
            self.dump()
        else:
            index = pd.MultiIndex(
                names=["project_id", "sample_id"], levels=[[], []], codes=[[], []]
            )
            self.df = pd.DataFrame(
                columns=self.project_df_default_values.keys(), index=index
            )

        self.logger = logging.getLogger(logger_name)

    def compute_max_barcode_mismatch(self, indices: List[str]) -> int:
        """compute_max_barcode_mismatch.

        :param indices: List of illumina I7 index barcodes
        :type indices: List[str]
        :return: the maximum mismatch to be allowed for this set of index barcodes
        :rtype: int
        """
        num_samples = len(indices)

        if num_samples == 1:
            return 4
        else:
            max_mismatch = 3
            for i in range(num_samples - 1):
                for j in range(i + 1, num_samples):
                    hd = self.hamming_distance(indices[i], indices[j])
                    max_mismatch = min(max_mismatch, math.ceil(hd / 2) - 1)

        return max_mismatch

    def hamming_distance(self, string1: str, string2: str) -> int:
        """Cacluate hamming distance between two strings

        :param string1:
        :type string1: str
        :param string2:
        :type string2: str
        :rtype: int
        """
        return sum(c1 != c2 for c1, c2 in zip(string1, string2))

    def find_barcode_file(self, puck_id: str) -> str:
        """Tries to find path of a barcode file, using the puck_id.

        :param puck_id: puck_id of the puck we are looking for.
        :type puck_id: str
        :return: path of the puck file, containing barcodes, or None
        :rtype: str
        """
        # first find directory of puck file

        # return none or the path of the file
        def get_barcode_file(path):
            """get_barcode_file.

            :param path:
            """
            if os.path.isfile(path):
                return path

            return None

        def find_dir(name, path):
            """find_dir.

            :param name:
            :param path:
            """
            for root, dirs, files in os.walk(path):
                if name in dirs:
                    return os.path.join(root, name)

        puck_dir = find_dir(puck_id, self.config.puck_data["root"])
        path = None

        if puck_dir is not None:
            # puck dir exists, look for barcode file pattern
            path = os.path.join(puck_dir, self.config.puck_data["barcode_file"])

            return get_barcode_file(path)
        else:
            return self.project_df_default_values["puck_barcode_file"]

    def get_sample_info(self, project_id: str, sample_id: str) -> Dict:
        """get_sample_info.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :return: A dictionary containing all the values of a given sample, from the ProjectDF.
        :rtype: Dict
        """
        # returns sample info from the projects df
        out_dict = self.df.loc[(project_id, sample_id)].to_dict()

        return out_dict

    def is_external(self, project_id: str, sample_id: str) -> bool:
        """is_external.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)

        data = self.df.loc[
            (project_id, sample_id),
            [
                "R1",
                "R2",
                "basecalls_dir",
                "sample_sheet",
                "longreads",
                "dge",
                "is_merged",
            ],
        ]

        if (
            (data.R1 and data.R2)
            or (data.basecalls_dir and data.sample_sheet)
            or (data.longreads)
            and not data.dge
            or data.is_merged
        ):
            return False
        elif data.dge:
            return True
        else:
            raise SpacemakeError(f'Sample with id (project_id, sample_id)=' +
                f'({project_id}, {sample_id}) is invalid.')

    def has_dge(self, project_id: str, sample_id: str) -> bool:
        """Returns True if a has dge. for Pacbio only samples returns False.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)

        data = self.df.loc[
            (project_id, sample_id),
            [
                "R1",
                "R2",
                "basecalls_dir",
                "sample_sheet",
                "longreads",
                "dge",
                "is_merged",
            ],
        ]

        if (
            data.is_merged
            or (data.R1 and data.R2)
            or (data.sample_sheet and data.basecalls_dir)
            or data.dge
        ):
            return True
        elif data.longreads:
            return False
        else:
            raise SpacemakeError(f'Sample with id (project_id, sample_id)=' +
                f'({project_id}, {sample_id}) is invalid.')
            

    def is_spatial(self, project_id: str, sample_id: str) -> bool:
        """Returns true if a sample with index (project_id, sample_id) is spatial,
        meaning that it has spatial barcodes attached.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)
        puck_barcode_file = self.get_metadata(
            "puck_barcode_file", project_id=project_id, sample_id=sample_id
        )

        puck_name = self.get_metadata(
            "puck", project_id=project_id, sample_id=sample_id
        )

        puck = self.config.get_puck(puck_name, return_empty=True)

        if puck_barcode_file is not None or puck.has_barcodes:
            return True
        else:
            return False

    def get_puck_variables(
        self, project_id: str, sample_id: str, return_empty=False
    ) -> Dict:
        """get_puck_variables.

        :param project_id: project_id of a sample
        :type project_id: str
        :param sample_id: sample_id of a sample
        :type sample_id: str
        :param return_empty:
        :return: A dictionary containing the puck variables of a given sample
        :rtype: Dict
        """
        puck_name = self.get_metadata(
            "puck", project_id=project_id, sample_id=sample_id
        )

        return self.config.get_puck(puck_name, return_empty=return_empty).variables

    def get_metadata(self, field, project_id=None, sample_id=None, **kwargs):
        """get_metadata.

        :param field:
        :param project_id:
        :param sample_id:
        :param kwargs:
        """
        df = self.df
        if sample_id is not None:
            df = df.query("sample_id == @sample_id")

        if project_id is not None:
            df = df.query("project_id == @project_id")

        for key, value in kwargs.items():
            df = df.loc[df.loc[:, key] == value]

        dl = df.loc[:, field].to_list()
        if len(dl):
            return dl[0]
        else:
            return ""

    def dump(self):
        """dump."""
        self.df.to_csv(self.file_path)

    def add_sample_sheet(self, sample_sheet_path, basecalls_dir):
        """add_sample_sheet.

        :param sample_sheet_path:
        :param basecalls_dir:
        """
        with open(sample_sheet_path) as sample_sheet:
            ix = 0
            investigator = None
            sequencing_date = None

            for line in sample_sheet:
                line = line.strip("\n")
                if "Investigator" in line:
                    investigator = line.split(",")[1]
                if "Date" in line:
                    sequencing_date = line.split(",")[1]
                if "[Data]" in line:
                    # the counter ix stops here
                    break
                else:
                    ix = ix + 1

        # read everything after [Data]
        df = pd.read_csv(sample_sheet_path, skiprows=ix + 1)
        # rename columns
        to_rename = {
            "Sample_ID": "sample_id",
            "Sample_Name": "puck_id",
            "Sample_Project": "project_id",
            "Description": "experiment",
            "index": "index",
        }
        df.rename(
            columns=to_rename,
            inplace=True,
        )
        # select only renamed columns
        df = df[to_rename.values()]
        df["species"] = df["experiment"].str.split("_").str[-1]
        df["investigator"] = investigator
        df["sequencing_date"] = sequencing_date

        # rename columns
        df["basecalls_dir"] = basecalls_dir
        df["demux_barcode_mismatch"] = self.compute_max_barcode_mismatch(df["index"])
        df["sample_sheet"] = sample_sheet_path
        df["demux_dir"] = (
            df["sample_sheet"].str.split("/").str[-1].str.split(".").str[0]
        )
        df["puck_barcode_file"] = df.puck_id.apply(self.find_barcode_file)
        df.set_index(["project_id", "sample_id"], inplace=True)

        for ix, row in df.iterrows():
            self.add_update_sample(project_id=ix[0], sample_id=ix[1], **row.to_dict())

    def assert_index_value(self, index_value, index_level):
        if not isinstance(index_value, list):
            index_value = [index_value]

        ixs = self.df.index.get_level_values(index_level)

        for ixv in index_value:
            if ixv not in ixs:
                raise ProjectSampleNotFoundError(index_level, ixv)

    def sample_exists(self, project_id=None, sample_id=None):
        """sample_exists.

        :param project_id:
        :param sample_id:
        """
        if project_id is None or sample_id is None:
            raise Exception(
                f"you need to provide a sample_id and project_id to check if sample exists"
            )
        else:
            ix = (project_id, sample_id)

            return ix in self.df.index

    def assert_sample(self, project_id, sample_id):
        if not self.sample_exists(project_id, sample_id):
            raise ProjectSampleNotFoundError(
                "(project_id, sample_id)", (project_id, sample_id)
            )

    def add_update_sample(
        self,
        action=None,
        project_id=None,
        sample_id=None,
        R1=None,
        R2=None,
        dge=None,
        longreads=None,
        longread_signature=None,
        sample_sheet=None,
        basecalls_dir=None,
        is_merged=False,
        return_series=False,
        **kwargs,
    ):
        """add_update_sample.

        :param action:
        :param project_id:
        :param sample_id:
        :param R1:
        :param R2:
        :param dge:
        :param longreads:
        :param longread_signature:
        :param sample_sheet:
        :param basecalls_dir:
        :param is_merged:
        :param return_series:
        :param kwargs:
        """
        ix = (project_id, sample_id)
        sample_exists = self.sample_exists(*ix)

        if action is None and sample_exists:
            action = "update"
        elif action is None and not sample_exists:
            action = "add"
        elif action == "add" and sample_exists:
            raise SampleAlreadyExistsError(ix)
        elif action == "update" and not sample_exists:
            raise ProjectSampleNotFoundError("(project_id, sample_id)", ix)

        if action == "add":
            self.logger.info(f"Adding sample {ix}")
        elif action == "update":
            self.logger.info(f"Updating sample {ix}")
        else:
            raise ValueError(f"Unknown action {action}")

        # check variables
        # First we check if R1 and R2 is present.
        # If not, we check for longreads. If provided, we also need
        #   --longread-signature
        # If longreads not provided, we try with basecalls_dir and sample_sheet
        #   (only used by add_sample_sheet command)
        # If those area also not provided, we try to add a simple dge
        if action == "add" and (R1 is None or R2 is None) and not is_merged:
            self.logger.info("R1 or R2 not provided, trying longreads")

            if not longreads:
                self.logger.info(
                    "longreads not provided, trying basecalls_dir and sample_sheet"
                )
                if basecalls_dir is None or sample_sheet is None:
                    self.logger.info(
                        "basecalls_dir or sample_sheet not provided, trying dge"
                    )
                    if not dge:
                        raise SpacemakeError(
                            "Neither R1 & R2, longreads, basecalls_dir & "
                            + "sample_sheet, nor dge were provided.\n"
                            + "Some reads/data has to be provided"
                        )
            else:
                if not longread_signature:
                    raise SpacemakeError(
                        "adding longreads requires to set --longread-signature as well (e.g. dropseq, noUMI, default, visium, slideseq_bc14,...)"
                    )

        # assert files first
        assert_file(R1, default_value=None, extension=".fastq.gz")
        assert_file(R2, default_value=None, extension=".fastq.gz")
        assert_file(longreads, default_value=None, extension="all")
        assert_file(
            dge,
            default_value=None,
            extension=[".h5", ".csv", ".h5ad", ".loom", ".txt", ".txt.gz"],
        )

        # assign reads
        # if there are strings, make them lists, as one sample can have many read files
        if R1 is not None and isinstance(R1, str):
            R1 = [R1]

        if R2 is not None and isinstance(R2, str):
            R2 = [R2]

        if R1 is not None and R2 is not None:
            if len(R1) != len(R2):
                raise SpacemakeError(
                    f"Trying to set an unmatching number of "
                    + f"read pairs for sample: {ix}.\n"
                    + f"# of R1 files = {len(R1)}\n"
                    + f"# of R2 files = {len(R2)}"
                )

        is_spatial = assert_file(
            kwargs.get(
                "puck_barcode_file", self.project_df_default_values["puck_barcode_file"]
            ),
            default_value=self.project_df_default_values["puck_barcode_file"],
        )

        if "run_mode" in kwargs and isinstance(kwargs["run_mode"], str):
            # if a single run mode is provided as a string
            # create a list manually
            kwargs["run_mode"] = [kwargs["run_mode"]]

        # check if run mode exists
        for run_mode in kwargs.get("run_mode", []):
            if not self.config.variable_exists("run_modes", run_mode):
                raise ConfigVariableNotFoundError("run_modes", run_mode)

        config_variables_to_check = {
            "pucks": "puck",
            "barcode_flavors": "barcode_flavor",
            "species": "species",
        }

        for cv_plural, cv_singular in config_variables_to_check.items():
            if cv_singular in kwargs.keys():
                if not self.config.variable_exists(cv_plural, kwargs[cv_singular]):
                    raise ConfigVariableNotFoundError(cv_singular, kwargs[cv_singular])

        # if everything correct, add or update
        # first populate kwargs
        kwargs["R1"] = R1
        kwargs["R2"] = R2
        kwargs["dge"] = dge
        kwargs["longreads"] = longreads
        kwargs["longread_signature"] = longread_signature
        kwargs["sample_sheet"] = sample_sheet
        kwargs["basecalls_dir"] = basecalls_dir
        kwargs["is_merged"] = is_merged

        if sample_exists:
            new_project = self.df.loc[ix].copy()
            # pd.Series.update will only update values which are not None
            new_project.update(pd.Series(kwargs))
            self.df.loc[ix] = new_project
        else:
            # if sample is spatial, and puck not provided, assign 'default'
            if is_spatial and "puck" not in kwargs:
                kwargs["puck"] = "default"

            new_project = pd.Series(self.project_df_default_values)
            new_project.name = ix
            new_project.update(kwargs)

            self.df = self.df.append(new_project)

        if return_series:
            return (ix, new_project)

    def delete_sample(self, project_id, sample_id):
        """delete_sample.

        :param project_id:
        :param sample_id:
        """
        if self.sample_exists(project_id, sample_id):
            self.logger.info(
                f"Deleting sample: {ix}, with the following"
                + f" with the following variables:\n{element}"
            )

            ix = (project_id, sample_id)
            element = self.df.loc[ix]
            self.df.drop(ix, inplace=True)
            return element
        else:
            raise ProjectSampleNotFoundError(
                "(project_id, sample_id)", (project_id, sample_id)
            )

    def add_samples_from_yaml(self, projects_yaml_file):
        """add_samples_from_yaml.

        :param projects_yaml_file:
        """
        config = yaml.load(open(projects_yaml_file), Loader=yaml.FullLoader)
        demux_projects = config.get("projects", None)

        if demux_projects is not None:
            # if we have projects in the config file
            # get the samples
            for ip in demux_projects:
                self.add_sample_sheet(ip["sample_sheet"], ip["basecalls_dir"])

        # add additional samples from config.yaml, which have already been demultiplexed.
        for project in config["additional_projects"]:
            self.add_update_sample(**project)

    def get_ix_from_project_sample_list(self, project_id_list=[], sample_id_list=[]):
        """get_ix_from_project_sample_list.

        :param project_id_list:
        :param sample_id_list:
        """

        # raise error if both lists are empty
        if project_id_list == [] and sample_id_list == []:
            raise NoProjectSampleProvidedError()

        # of only one provided use that, if both use intersection
        if project_id_list == []:
            ix = self.df.query("sample_id in @sample_id_list").index
        elif sample_id_list == []:
            ix = self.df.query("project_id in @project_id_list").index
        else:
            ix = self.df.query(
                "project_id in @project_id_list and sample_id in @sample_id_list"
            ).index

        return ix

    def set_variable(self, ix, variable_name, variable_key, keep_old=False):
        """set_variable.

        :param ix:
        :param variable_name:
        :param variable_key:
        :param keep_old:
        """
        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)
        self.config.assert_variable(variable_name_pl, variable_key)

        # get current value
        i_variable = self.df.at[ix, variable_name]

        # if list, we either append or not
        if isinstance(i_variable, list):
            if keep_old:
                # keep the old list as well
                i_variable = list(set(i_variable + variable_key))
            else:
                i_variable = list(set(variable_key))

        # if we do not keep the old, simply set
        else:
            i_variable = variable_key

        if i_variable == [] or i_variable is None:
            raise EmptyConfigVariableError(variable_name)

        self.df.at[ix, variable_name] = i_variable

    def remove_variable(self, ix, variable_name, variable_key):
        """remove_variable.

        :param ix:
        :param variable_name:
        :param variable_key:
        """
        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)
        self.config.assert_variable(variable_name_pl, variable_key)

        if not isinstance(variable_key, list):
            raise ValueError("variable_key has to be a list")

        i_variable = self.df.at[ix, variable_name]

        i_variable = [val for val in i_variable if val not in variable_key]
        if i_variable == [] or i_variable is None:
            raise EmptyConfigVariableError(variable_name)

        self.df.at[ix, variable_name] = i_variable

    def set_remove_variable(
        self,
        variable_name,
        variable_key,
        action,
        project_id_list=[],
        sample_id_list=[],
        keep_old=False,
    ):
        """set_remove_variable.

        :param variable_name:
        :param variable_key:
        :param action:
        :param project_id_list:
        :param sample_id_list:
        :param keep_old:
        """
        self.assert_projects_samples_exist(project_id_list, sample_id_list)

        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)

        ix = self.get_ix_from_project_sample_list(
            project_id_list=project_id_list, sample_id_list=sample_id_list
        )

        if isinstance(variable_key, list):
            for var in variable_key:
                self.config.assert_variable(variable_name_pl, var)
        else:
            self.config.assert_variable(variable_name_pl, variable_key)

        for i, row in self.df.loc[ix, :].iterrows():
            # add/remove variable. if it already exists dont do anything
            if action == "set":
                self.set_variable(i, variable_name, variable_key, keep_old)
            elif action == "remove":
                self.remove_variable(i, variable_name, variable_key)

        return ix.to_list()

    def assert_projects_samples_exist(self, project_id_list=[], sample_id_list=[]):
        """assert_projects_samples_exist.

        :param project_id_list:
        :param sample_id_list:
        """
        for project in project_id_list:
            if project not in self.df.index.get_level_values("project_id"):
                raise ProjectSampleNotFoundError("project_id", project)

        for sample in sample_id_list:
            if sample not in self.df.index.get_level_values("sample_id"):
                raise ProjectSampleNotFoundError("sample_id", sample)

    def merge_samples(
        self,
        merged_project_id,
        merged_sample_id,
        project_id_list=[],
        sample_id_list=[],
        **kwargs,
    ):
        """merge_samples.

        :param merged_project_id:
        :param merged_sample_id:
        :param project_id_list:
        :param sample_id_list:
        :param kwargs:
        """
        # check if projects and samples with these IDs exist
        self.assert_projects_samples_exist(project_id_list, sample_id_list)

        ix = self.get_ix_from_project_sample_list(
            project_id_list=project_id_list, sample_id_list=sample_id_list
        )

        consistent_variables = list(self.config.main_variables_sg2pl.keys())
        consistent_variables.remove("run_mode")

        ix_list = ix.to_list()

        if ix_list == []:
            raise ProjectSampleNotFoundError(
                "(project_id_list, sample_id_list)", (project_id_list, sample_id_list)
            )

        # check for variable inconsistency
        # raise error if variable different between samples
        for variable in consistent_variables:
            if (variable in kwargs and
                variable not in ['species', 'barcode_flavor']):
                # if variable provided from command line, skip
                self.logger.info(f'{variable} provided, skipping deduction...')
                continue
            
            self.logger.info(f'{variable} not provided. deducing from merged samples...')

            variable_val = self.df.loc[ix, variable].to_list()

            if len(set(variable_val)) > 1:
                raise InconsistentVariablesDuringMerge(
                    variable_name=variable, variable_value=variable_val, ix=ix
                )
            else:
                # attach the deduced, consisten variable
                kwargs[variable] = variable_val[0]

        # get puck_barcode_file
        if "puck_barcode_file" not in kwargs:
            pbf_default = self.project_df_default_values["puck_barcode_file"]
            pbf_list = self.df.loc[ix, "puck_barcode_file"].to_list()
            # filter out default values
            pbf_list = [x for x in pbf_list if x != pbf_default]

            # remove duplicates
            pbf_list = list(set(pbf_list))

            if pbf_list == []:
                # if all values are default
                kwargs["puck_barcode_file"] = pbf_default
            elif len(pbf_list) == 1:
                kwargs["puck_barcode_file"] = pbf_list[0]
            else:
                raise InconsistentVariablesDuringMerge(
                    variable_name="puck_barcode_file",
                    variable_value=pbf_list,
                    ix=ix.to_list(),
                )

        # after all checks, log that we are merging
        self.logger.info(f'Merging samples {ix_list} together\n')
        variables_to_deduce = ["investigator", "experiment", "sequencing_date"]

        for variable in variables_to_deduce:
            if variable not in kwargs:
                kwargs[variable] = ";".join(self.df.loc[ix, variable].unique())

        # if no run_mode provided, overwrite with user defined one
        if "run_mode" not in kwargs.keys():
            run_mode_lists = [
                set(run_mode) for run_mode in self.df.loc[ix].run_mode.to_list()
            ]

            # join run modes from parent samples
            if len(run_mode_lists) == 1:
                run_mode = run_mode_lists[0]
            else:
                run_mode = run_mode_lists[0].intersection(*run_mode_lists[1:])

            # create a list from the set intersection
            run_mode = list(run_mode)

            # if there are no common elements, throw an error
            if len(run_mode) == 0:
                raise InconsistentVariablesDuringMerge(
                    variable_name="run_mode", variable_value=run_mode_lists, ix=ix_list
                )

            # finally add run mode to arguments
            kwargs["run_mode"] = run_mode

        # set the action to add
        kwargs["action"] = "add"

        sample_added = self.add_update_sample(
            project_id=merged_project_id,
            sample_id=merged_sample_id,
            is_merged=True,
            merged_from=ix.to_list(),
            **kwargs,
        )

        return (sample_added, ix)
